from functools import partial

import tpDcc
from tpDcc.libs.qt.core import mixin
from tpDcc.libs.qt.widgets import layouts, toast, buttons, dividers


class ToastExample(tpDcc.Window, mixin.FieldMixin):
    def __init__(self):
        super(ToastExample, self).__init__(title='Example for Toast')

    def ui(self):
        super(ToastExample, self).ui()

        btn1 = buttons.BaseButton('Normal Message').primary()
        btn2 = buttons.BaseButton('Success Message').success()
        btn3 = buttons.BaseButton('Warning Message').warning()
        btn4 = buttons.BaseButton('Error Message').danger()

        btn1.clicked.connect(partial(self._on_show_toast, toast.BaseToast.info, {'text': 'Normal'}))
        btn2.clicked.connect(partial(self._on_show_toast, toast.BaseToast.success, {'text': 'Success'}))
        btn3.clicked.connect(partial(self._on_show_toast, toast.BaseToast.warning, {'text': 'Warning'}))
        btn4.clicked.connect(partial(self._on_show_toast, toast.BaseToast.error, {'text': 'Error'}))

        sub_lyt1 = layouts.HorizontalLayout()
        sub_lyt1.addWidget(btn1)
        sub_lyt1.addWidget(btn2)
        sub_lyt1.addWidget(btn3)
        sub_lyt1.addWidget(btn4)

        loading_btn = buttons.BaseButton('Loading Toast').primary()
        loading_btn.clicked.connect(self._on_show_loading_toast)

        self.main_layout.addWidget(dividers.Divider('different type'))
        self.main_layout.addLayout(sub_lyt1)
        self.main_layout.addWidget(dividers.Divider('loading'))
        self.main_layout.addWidget(loading_btn)
        self.main_layout.addStretch()

    def _on_show_toast(self, fn, config):
        fn(parent=self, **config)

    def _on_show_loading_toast(self):
        msg = toast.BaseToast.loading('Loading', parent=self)
        msg.toastClosed.connect(partial(toast.BaseToast.success, 'Loading completed!', self))


def run():
    return ToastExample()



from Qt.QtGui import *

import tpDcc
from tpDcc.libs.qt.core import base, mixin
from tpDcc.libs.qt.widgets import layouts, panel, label, buttons, group, dividers


class SliderPanelExample(tpDcc.Window, mixin.FieldMixin):
    def __init__(self):
        super(SliderPanelExample, self).__init__(title='Example for Slider Panel')

    def ui(self):
        super(SliderPanelExample, self).ui()

        add_icon = tpDcc.ResourcesMgr().icon('add', color=QColor(125, 0, 0))

        self._button_grp = group.RadioButtonGroup()
        self._button_grp.set_button_list(['top', {'text': 'right', 'checked': True}, 'bottom', 'left'])

        self._open_btn_2 = buttons.BaseButton('Open', parent=self).tiny()
        lyt = layouts.HorizontalLayout()
        lyt.addWidget(self._button_grp)
        lyt.addSpacing(20)
        lyt.addWidget(self._open_btn_2)
        lyt.addStretch()

        self._new_account_btn = buttons.BaseButton(text='New account', icon=add_icon).primary()
        lyt1 = layouts.HorizontalLayout()
        lyt1.addWidget(label.BaseLabel('Submit form in drawer'))
        lyt1.addWidget(self._new_account_btn)
        lyt1.addStretch()

        self.main_layout.addWidget(dividers.Divider('Custom Placement'))
        self.main_layout.addLayout(lyt)
        self.main_layout.addWidget(dividers.Divider('Submit form in drawer'))
        self.main_layout.addLayout(lyt1)

    def setup_signals(self):
        self._open_btn_2.clicked.connect(self._on_open_button_2)
        self._new_account_btn.clicked.connect(self._on_new_account)

    def _on_open_button_2(self):
        custom_widget = base.BaseWidget()
        custom_widget.main_layout.addWidget(label.BaseLabel('Some contentes ...'))
        custom_widget.main_layout.addWidget(label.BaseLabel('Some contentes ...'))
        custom_widget.main_layout.addWidget(label.BaseLabel('Some contentes ...'))

        slider_panel = panel.SliderPanel('Basic Panel', parent=self)
        slider_panel.position = self._button_grp.get_button_group().checkedButton().text()
        slider_panel.setFixedWidth(200)
        slider_panel.set_widget(custom_widget)
        slider_panel.show()

    def _on_new_account(self):
        pass


def run():
    return SliderPanelExample()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains functions related with Qt drop behaviour
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import Qt, QPoint, QRect, QRectF, QLineF, QSizeF
from Qt.QtWidgets import QBoxLayout, QWidget, QFrame, QLabel
from Qt.QtGui import QCursor, QPixmap, QColor, QPalette, QPainter, QBrush, QLinearGradient

from tpDcc.libs.qt.core import qtutils
from tpDcc.libs.qt.widgets import layouts


class DropArea(object):
    InvalidDropArea = 0
    TopDropArea = 1
    RightDropArea = 2
    BottomDropArea = 3
    LeftDropArea = 4
    CenterDropArea = 5

    AllAreas = [TopDropArea, RightDropArea, BottomDropArea, LeftDropArea, CenterDropArea]


class DropOverlay(QFrame, object):
    """
    Paints a translucent rectangle over another widget. The geometry of the rectangle
    is based on the mouse location
    """
    def __init__(self, parent=None):
        super(DropOverlay, self).__init__(parent=parent)

        self._allowed_areas = [DropArea.InvalidDropArea]
        self._cross = DropOverlayCross(overlay=self)
        self._full_area_drop = False
        self._last_location = DropArea.InvalidDropArea
        self._target = None
        self._target_rect = QRect()

        self.setWindowFlags(Qt.Tool | Qt.FramelessWindowHint)
        self.setWindowOpacity(0.2)
        self.setWindowTitle('DropOverlay')

        top_bottom_layout = QBoxLayout(QBoxLayout.TopToBottom)
        top_bottom_layout.setContentsMargins(0, 0, 0, 0)
        top_bottom_layout.setSpacing(0)
        self.setLayout(top_bottom_layout)

        area_widgets = dict()
        area_widgets[DropArea.TopDropArea] = self.create_drop_indicator_widget(DropArea.TopDropArea)
        area_widgets[DropArea.RightDropArea] = self.create_drop_indicator_widget(DropArea.RightDropArea)
        area_widgets[DropArea.BottomDropArea] = self.create_drop_indicator_widget(DropArea.BottomDropArea)
        area_widgets[DropArea.LeftDropArea] = self.create_drop_indicator_widget(DropArea.LeftDropArea)
        area_widgets[DropArea.CenterDropArea] = self.create_drop_indicator_widget(DropArea.CenterDropArea)
        self._cross.set_area_widgets(area_widgets)
        self._cross.setVisible(False)
        self.setVisible(False)

    # region Static Functions
    @staticmethod
    def create_drop_indicator_pixmap(palette, size, drop_area):
        border_color = palette.color(QPalette.Active, QPalette.Highlight)
        background_color = palette.color(QPalette.Active, QPalette.Base)
        area_background_color = palette.color(QPalette.Active, QPalette.Highlight).lighter(150)

        pm = QPixmap(size.width(), size.height())
        pm.fill(QColor(0, 0, 0, 0))

        painter = QPainter(pm)
        pen = painter.pen()
        base_rect = QRectF(pm.rect())

        painter.fillRect(base_rect, background_color)

        painter.save()
        area_rect = QRectF()
        area_line = QLineF()
        gradient = QLinearGradient()
        if drop_area == DropArea.TopDropArea:
            area_rect = QRectF(base_rect.x(), base_rect.y(), base_rect.width(), base_rect.height() * 0.5)
            area_line = QLineF(area_rect.bottomLeft(), area_rect.bottomRight())
            gradient.setStart(area_rect.topLeft())
            gradient.setFinalStop(area_rect.bottomLeft())
            gradient.setColorAt(0, area_background_color)
            gradient.setColorAt(1, area_background_color.lighter(120))
        elif drop_area == DropArea.RightDropArea:
            area_rect = QRectF(base_rect.width() * 0.5, base_rect.y(), base_rect.width() * 0.5, base_rect.height())
            area_line = QLineF(area_rect.topLeft(), area_rect.bottomLeft())
            gradient.setStart(area_rect.topLeft())
            gradient.setFinalStop(area_rect.topRight())
            gradient.setColorAt(0, area_background_color.lighter(120))
            gradient.setColorAt(1, area_background_color)
        elif drop_area == DropArea.BottomDropArea:
            area_rect = QRectF(base_rect.x(), base_rect.height() * 0.5, base_rect.width(), base_rect.height() * 0.5)
            area_line = QLineF(area_rect.topLeft(), area_rect.topRight())
            gradient.setStart(area_rect.topLeft())
            gradient.setFinalStop(area_rect.bottomLeft())
            gradient.setColorAt(0, area_background_color.lighter(120))
            gradient.setColorAt(1, area_background_color)
        elif drop_area == DropArea.LeftDropArea:
            area_rect = QRectF(base_rect.x(), base_rect.y(), base_rect.width() * 0.5, base_rect.height())
            area_line = QLineF(area_rect.topRight(), area_rect.bottomRight())
            gradient.setStart(area_rect.topLeft())
            gradient.setFinalStop(area_rect.topRight())
            gradient.setColorAt(0, area_background_color)
            gradient.setColorAt(1, area_background_color.lighter(120))

        if area_rect.isValid():
            painter.fillRect(area_rect, gradient)
            pen = painter.pen()
            pen.setColor(border_color)
            pen.setStyle(Qt.DashLine)
            painter.setPen(pen)
            painter.drawLine(area_line)

        painter.restore()

        painter.save()
        pen = painter.pen()
        pen.setColor(border_color)
        pen.setWidth(1)

        painter.setPen(pen)
        painter.drawRect(base_rect.adjusted(0, 0, -pen.width(), -pen.width()))
        painter.restore()

        return pm

    @staticmethod
    def create_drop_indicator_widget(drop_area):
        layout = QLabel()
        layout.setObjectName('DropAreaLabel')
        metric = (layout.fontMetrics().height()) * 2.0
        size = QSizeF(metric, metric)
        layout.setPixmap(
            DropOverlay.create_drop_indicator_pixmap(palette=layout.palette(), size=size, drop_area=drop_area))

        return layout
    # endregion

    # region Override Functions
    def paintEvent(self, event):
        painter = QPainter(self)
        area_color = self.palette().color(QPalette.Active, QPalette.Highlight)

        if self._full_area_drop:
            r = self.rect()
            painter.fillRect(r, QBrush(area_color, Qt.Dense4Pattern))
            painter.setBrush(QBrush(area_color))
            painter.drawRect(r)
            return

        r = self.rect()
        drop_area = self.cursor_location()
        if drop_area == DropArea.TopDropArea:
            r.setHeight(r.height() * 0.5)
        elif drop_area == DropArea.RightDropArea:
            r.setX(r.width() * 0.5)
        elif drop_area == DropArea.BottomDropArea:
            r.setY(r.height() * 0.5)
        elif drop_area == DropArea.LeftDropArea:
            r.setWidth(r.width() * 0.5)
        elif drop_area == DropArea.CenterDropArea:
            r = self.rect()

        if not r.isNull():
            painter.fillRect(r, QBrush(area_color, Qt.Dense4Pattern))
            painter.setBrush(QBrush(area_color))
            painter.drawRect(r)

        # Draw rect over the entire size + border
        # r = self.rect()
        # r.setWidth(r.width() - 1)
        # r.setHeight(r.height() - 1)
        # painter.fillRect(r, QBrush(QColor(0, 100, 255), Qt.Dense4Pattern))
        # painter.setBrush(QColor(0, 100, 255))
        # painter.drawRect(r)

    def showEvent(self, event):
        self._cross.show()

    def hideEvent(self, event):
        self._cross.hide()

    def resizeEvent(self, event):
        self._cross.resize(event.size())

    def moveEvent(self, event):
        self._cross.move(event.pos())
        self._cross.move(event.pos())
    # endregion

    # region Public Functions
    def set_allowed_areas(self, areas):
        if areas == self._allowed_areas:
            return
        self._allowed_areas = areas
        self._cross.reset()

    def allowed_areas(self):
        return self._allowed_areas

    def set_area_widgets(self, widgets):
        self._cross.set_area_widgets(widgets=widgets)

    def cursor_location(self):
        return self._cross.cursor_location()

    def show_drop_overlay(self, target, target_area_rect=None):

        if target_area_rect is None:
            if self._target == target:
                drop_area = self.cursor_location()
                if drop_area != self._last_location:
                    self.repaint()
                    self._last_location = drop_area
                return drop_area

            self.hide_drop_overlay()
            self._full_area_drop = False
            self._target = target
            self._target_rect = QRect()
            self._last_location = DropArea.InvalidDropArea

            self.resize(target.size())
            self.move(target.mapToGlobal(target.rect().topLeft()))

            self.show()

            return self.cursor_location()
        else:
            if self._target == target and self._target_rect == target_area_rect:
                return

            self.hide_drop_overlay()
            self._full_area_drop = True
            self._target = target
            self._target_rect = target_area_rect
            self._last_location = DropArea.InvalidDropArea

            self.resize(target_area_rect.size())
            self.move(target.mapToGlobal(QPoint(target_area_rect.x(), target_area_rect.y())))

            self.show()

            return

    def hide_drop_overlay(self):
        self.hide()
        self._full_area_drop = False

        # Check if Qt Version > 5.0.0 -> If True:
        # self._target.clear
        # else
        # self._target = 0

        self._target = 0
        self._target_rect = QRect()
        self._last_location = DropArea.InvalidDropArea
    # endregion


class DropOverlayCross(QWidget, object):
    """
    Shows a cross with 5 different drop area possibilities
    """
    def __init__(self, overlay):
        super(DropOverlayCross, self).__init__(overlay.parentWidget())

        self._overlay = overlay
        self._widgets = None

        self.setWindowFlags(Qt.Tool | Qt.FramelessWindowHint)
        self.setWindowTitle('DropOverlayCross')
        # self.setAttribute(Qt.WA_TranslucentBackground)

        self._grid = layouts.GridLayout(spacing=0, margins=(0, 0, 0, 0))

        bl1 = QBoxLayout(QBoxLayout.TopToBottom)
        bl1.setContentsMargins(0, 0, 0, 0)
        bl1.setSpacing(0)
        self.setLayout(bl1)

        bl2 = QBoxLayout(QBoxLayout.LeftToRight)
        bl2.setContentsMargins(0, 0, 0, 0)
        bl2.setSpacing(0)

        bl1.addStretch(1)
        bl1.addLayout(bl2)
        bl2.addStretch(1)
        bl2.addLayout(self._grid, 0)
        bl2.addStretch(1)
        bl1.addStretch(1)

    # region Static Functions
    @staticmethod
    def grid_pos_for_area(drop_area):
        """
        Given a drop area return the position in which the cross widget should be located
        :param drop_area: DropArea
        :return: tuple(QPoint, int)
        """

        if drop_area == DropArea.TopDropArea:
            return QPoint(0, 1), (Qt.AlignHCenter | Qt.AlignBottom)
        elif drop_area == DropArea.RightDropArea:
            return QPoint(1, 2), (Qt.AlignLeft | Qt.AlignVCenter)
        elif drop_area == DropArea.BottomDropArea:
            return QPoint(2, 1), (Qt.AlignHCenter | Qt.AlignTop)
        elif drop_area == DropArea.LeftDropArea:
            return QPoint(1, 0), (Qt.AlignRight | Qt.AlignVCenter)
        elif drop_area == DropArea.CenterDropArea:
            return QPoint(1, 1), Qt.AlignCenter
        else:
            return QPoint(), int()
    # endregion

    # region Override Functions
    def showEvent(self, event):
        self.resize(self._overlay.size())
        self.move(self._overlay.pos())
    # endregion

    # region Public Functions
    def set_area_widgets(self, widgets):
        qtutils.clear_layout(self._grid)
        self._widgets = widgets
        for drop_area, widget in self._widgets.items():
            opts = self.grid_pos_for_area(drop_area=drop_area)
            self._grid.addWidget(widget, opts[0].x(), opts[0].y(), opts[1])
        self.reset()

    def cursor_location(self):
        pos = self.mapFromGlobal(QCursor.pos())
        for drop_area, widget in self._widgets.items():
            if drop_area in self._overlay.allowed_areas():
                if widget:
                    if widget.isVisible() and widget.geometry().contains(pos):
                        return drop_area

        return DropArea.InvalidDropArea

    def reset(self):
        all_areas = [DropArea.TopDropArea, DropArea.RightDropArea, DropArea.BottomDropArea,
                     DropArea.LeftDropArea, DropArea.CenterDropArea]
        allowed_areas = self._overlay.allowed_areas()

        for i in range(len(all_areas)):
            opts = self.grid_pos_for_area(i)
            item = self._grid.itemAtPosition(opts[0].x(), opts[0].y())
            if item:
                w = item.widget()
                if w:
                    try:
                        w.setVisible(set(allowed_areas).issubset(set(all_areas)))
                    except Exception:
                        print('ALLOWED AREAS: ', allowed_areas)
                        print('ALL AREAS: ', all_areas)
    # endregion



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains classes to create breadcrumb widgets
"""

from __future__ import print_function, division, absolute_import

import os

from Qt.QtWidgets import QSizePolicy, QButtonGroup, QFrame

from tpDcc.libs.python import path, python
from tpDcc.libs.resources.core import theme
from tpDcc.libs.qt.core import mixin, base
from tpDcc.libs.qt.widgets import layouts, buttons, label


class Breadcrumb(object):
    def __init__(self, label):
        """
        Constructor
        :param label: QLabel, label used in this breadcrumb
        """

        self._label = label

    @property
    def label(self):
        return self._label


@theme.mixin
class BreadcrumbWidget(base.BaseWidget, object):
    """
    Widget that display current location withing a hierarchy
    It allows going back/forward inside a hierarchy
    """

    def __init__(self, separator=None, parent=None):
        super(BreadcrumbWidget, self).__init__(parent=parent)

        current_theme = self.theme()

        separator_color = current_theme.accent_color if current_theme else '#E2AC2C'

        self._separator = separator or "<span style='color:{}'> &#9656; </span>".format(separator_color)
        self._separators = list()

        self.setObjectName('BreadcrumbWidget')
        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

    def get_main_layout(self):
        main_layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        main_layout.addStretch()

        return main_layout

    def ui(self):
        super(BreadcrumbWidget, self).ui()

        self._button_group = QButtonGroup()

    def set_items(self, data_list):
        """
        Sets the the items of the breadcrumb cleaning old ones
        :param data_list:
        :return:
        """

        for btn in self._button_group.buttons():
            self._button_group.removeButton(btn)
            self.main_layout.removeWidget(btn)
            btn.setVisible(False)
            btn.deleteLater()
        for sep in self._separators:
            self.main_layout.removeWidget(sep)
            sep.setVisible(False)
            sep.deleteLater()
        python.clear_list(self._separators)

        for index, data_dict in enumerate(data_list):
            self.add_item(data_dict, index)

    def add_item(self, data_dict, index=None):
        """
        Adds a new item to the breadcrumb
        :param data_dict: dict
        :param index: int
        """

        btn = buttons.BaseToolButton()
        btn.setText(data_dict.get('text'))
        if data_dict.get('image'):
            btn.image(data_dict.get('image'))
        if data_dict.get('tooltip'):
            btn.setProperty('toolTip', data_dict.get('tooltip'))
        if data_dict.get('clicked'):
            btn.clicked.connect(data_dict.get('clicked'))
        if data_dict.get('text'):
            if data_dict.get('svg') or data_dict.get('icon'):
                btn.text_beside_icon()
            else:
                btn.text_only()
        else:
            btn.icon_only()

        if self._button_group.buttons():
            separator = label.BaseLabel(self._separator).secondary()
            self._separators.append(separator)
            self.main_layout.insertWidget(self.main_layout.count() - 1, separator)
        self.main_layout.insertWidget(self.main_layout.count() - 1, btn)

        if index is None:
            self._button_group.addButton(btn)
        else:
            self._button_group.addButton(btn, index)

    def set_from_path(self, file_path):
        """
        Creates a proper Breadcrumb list for given path and sets the text
        """

        self._widgets = list()
        file_path = os.path.dirname(file_path)
        folders = path.get_folders_from_path(file_path)
        data_list = list()
        for folder in folders:
            data_list.append({'text': folder})
        self.set_items(data_list)

    def get_breadcumbs(self):
        """
        Returns current list of breadcumb texts
        :return: list(str)
        """

        return [btn.text() for btn in self._button_group.buttons()]


class BreadcrumbFrame(QFrame, object):
    def __init__(self, separator=None, parent=None):
        super(BreadcrumbFrame, self).__init__(parent)

        self.setObjectName('TaskFrame')
        self.setFrameStyle(QFrame.StyledPanel)

        main_layout = layouts.HorizontalLayout(spacing=2, margins=(2, 2, 2, 2))
        self.setLayout(main_layout)

        self._breadcrumb = BreadcrumbWidget(separator=separator, parent=self)

        title_layout = layouts.HorizontalLayout()
        title_layout.addStretch()
        title_layout.addWidget(self._breadcrumb)
        title_layout.addStretch()

        main_layout.addLayout(title_layout)

    def set_items(self, data_list):
        """
        Sets the the items of the breadcrumb cleaning old ones
        :param data_list:
        :return:
        """

        self._breadcrumb.set_items(data_list)

    def set_from_path(self, file_path):
        """
        Creates a proper Breadcrumb list for given path and sets the text
        """

        self._breadcrumb.set_from_path(file_path)

    def get_breadcumbs(self):
        """
        Returns current
        :return: list(str)
        """

        return self._breadcrumb.get_breadcumbs()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains custom Qt slider widgets
"""

from __future__ import print_function, division, absolute_import

from copy import copy

from Qt.QtCore import Qt, Signal, Property, QPoint, QPointF, QEvent
from Qt.QtWidgets import QSizePolicy, QWidget, QAbstractSpinBox, QDoubleSpinBox, QStyle, QStyleOptionSlider
from Qt.QtWidgets import QGroupBox, QSlider, QToolTip, QMenu, QColorDialog
from Qt.QtGui import QCursor, QColor, QFont, QPainter, QBrush, QLinearGradient, QMouseEvent

from tpDcc import dcc
from tpDcc.libs.python import mathlib, color as core_color
from tpDcc.libs.resources.core import theme, color
from tpDcc.libs.qt.core import qtutils, contexts as qt_contexts
from tpDcc.libs.qt.widgets import layouts, label, buttons

FLOAT_SLIDER_DRAG_STEPS = [100.0, 10.0, 1.0, 0.1, 0.01, 0.001]
INT_SLIDER_DRAG_STEPS = [100.0, 10.0, 1.0]


@theme.mixin
class BaseSlider(QSlider, object):
    def __init__(self, orientation=Qt.Horizontal, parent=None):
        super(BaseSlider, self).__init__(orientation, parent=parent)

    def mouseMoveEvent(self, event):
        QToolTip.showText(event.globalPos(), str(self.value()), self)
        return super(BaseSlider, self).mouseMoveEvent(event)


class SliderDraggers(QWidget, object):
    increment = Signal(object)

    def __init__(self, parent=None, is_float=True, dragger_steps=None, main_color=None):
        super(SliderDraggers, self).__init__(parent)

        self._drags = list()
        self._initial_pos = None
        self._active_drag = None
        self._last_delta_x = 0
        self._change_direction = 0
        self._main_color = main_color if main_color else QColor(215, 128, 26).getRgb()
        dragger_steps = dragger_steps or FLOAT_SLIDER_DRAG_STEPS

        self.setWindowFlags(Qt.Popup)

        draggers_layout = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))
        self.setLayout(draggers_layout)

        steps = copy(dragger_steps)
        if not is_float:
            steps = list(filter(lambda x: abs(x) >= 1.0, steps))
        for i in steps:
            drag = HoudiniInputDragger(self, i)
            self._drags.append(drag)
            draggers_layout.addWidget(drag)

        self.installEventFilter(self)

    @property
    def active_drag(self):
        return self._active_drag

    @active_drag.setter
    def active_drag(self, input_dragger):
        self._active_drag = input_dragger

    @property
    def drags(self):
        return self._drags

    def eventFilter(self, obj, event):
        if event.type() == QEvent.MouseMove:
            if self._active_drag:
                modifiers = event.modifiers()
                self._active_drag.setStyleSheet(self._get_style_sheet())
                if not self._initial_pos:
                    self._initial_pos = event.globalPos()
                delta_x = event.globalPos().x() - self._initial_pos.x()
                self._change_direction = mathlib.clamp(delta_x - self._last_delta_x, -1.0, 1.0)
                if self._change_direction != 0:
                    v = self._change_direction * self._active_drag.factor
                    if modifiers == Qt.NoModifier and delta_x % 4 == 0:
                        self.increment.emit(v)
                    if modifiers in [Qt.ShiftModifier, Qt.ControlModifier] and delta_x % 8 == 0:
                        self.increment.emit(v)
                    if modifiers == Qt.ShiftModifier | Qt.ControlModifier and delta_x % 32 == 0:
                        self.increment.emit(v)
                self._last_delta_x = delta_x
        if event.type() == QEvent.MouseButtonRelease:
            self.hide()
            self._last_delta_x = 0
            del(self)

        return False

    def _get_style_sheet(self):
        return """
        QGroupBox{
            border: 0.5 solid darkgrey;
            background : %s;
            color: white;
        }
        QLabel{
            background: transparent;
            border: 0 solid transparent;
            color: white;
        }
        """ % "rgba%s" % str(self._main_color)


class Slider(QSlider, object):
    """
    Custom slider that allows:
    - Left/Mid: Click to move handle
    - Ctrl and drag to move handle half velocity
    - Shift and drag to move handle quarter velocity
    - Ctrl + Shift and drag to move handle eighty velocity
    """

    editingFinihsed = Signal()
    valueIncremented = Signal(object)
    floatValueChanged = Signal(object)

    def __init__(self, parent=None, dragger_steps=None, slider_range=None, *args, **kwargs):
        if dragger_steps is None:
            dragger_steps = INT_SLIDER_DRAG_STEPS
        if slider_range is None:
            slider_range = [-100, 100]
        super(Slider, self).__init__(parent=parent, **kwargs)

        self._slider_range = slider_range
        self._dragger_steps = dragger_steps
        self._is_float = False
        self._default_value = 0
        self._prev_value = 0
        self._delta_value = 0
        self._start_drag_pos = QPointF()
        self._real_start_drag_pos = QPointF()
        self._draggers = None

        if dcc.is_maya():
            self._left_button = Qt.MidButton
            self._mid_button = Qt.LeftButton
        else:
            self._left_button = Qt.LeftButton
            self._mid_button = Qt.MidButton

        self.setFocusPolicy(Qt.StrongFocus)
        self.setOrientation(Qt.Horizontal)
        self.setRange(self._slider_range[0], self._slider_range[1])

    def mousePressEvent(self, event):
        self._prev_value = self.value()
        self._start_drag_pos = event.pos()
        if event.button() == Qt.MidButton:
            if not self._draggers:
                self._draggers = SliderDraggers(parent=self, is_float=self._is_float, dragger_steps=self._dragger_steps)
                self._draggers.increment.connect(self.valueIncremented.emit)
            self._draggers.show()
            if self._is_float:
                self._draggers.move(
                    self.mapToGlobal(QPoint(event.pos().x() - 1, event.pos().y() - self._draggers.height() / 2)))
            else:
                draggers_height = self._draggers.height()
                self._draggers.move(
                    self.mapToGlobal(
                        QPoint(event.pos().x() - 1, event.pos().y() - (self._draggers.height() - draggers_height / 6))))
        elif event.button() == self._left_button and event.modifiers() not in \
                [Qt.ControlModifier, Qt.ShiftModifier, Qt.ControlModifier | Qt.ShiftModifier]:
            buttons = Qt.MouseButtons(self._mid_button)
            mouse_event = QMouseEvent(event.type(), event.pos(), self._mid_button, buttons, event.modifiers())
            super(Slider, self).mousePressEvent(mouse_event)
        elif event.modifiers() in [Qt.ControlModifier, Qt.ShiftModifier, Qt.ControlModifier | Qt.ShiftModifier]:
            style_slider = QStyleOptionSlider()
            style_slider.initFrom(self)
            style_slider.orientation = self.orientation()
            available = self.style().pixelMetric(QStyle.PM_SliderSpaceAvailable, style_slider, self)
            x_loc = QStyle.sliderPositionFromValue(
                self.minimum(), self.maximum(), super(Slider, self).value(), available)
            buttons = Qt.MouseButtons(self._mid_button)
            new_pos = QPointF()
            new_pos.setX(x_loc)
            mouse_event = QMouseEvent(event.type(), new_pos, self._mid_button, buttons, event.modifiers())
            self._start_drag_pos = new_pos
            self._real_start_drag_pos = event.pos()
            super(Slider, self).mousePressEvent(mouse_event)
            self._delta_value = self.value() - self._prev_value
            self.setValue(self._prev_value)
        else:
            super(Slider, self).mousePressEvent(event)

    def mouseMoveEvent(self, event):
        delta_x = event.pos().x() - self._real_start_drag_pos.x()
        delta_y = event.pos().y() - self._real_start_drag_pos.y()
        new_pos = QPointF()
        if event.modifiers() in [Qt.ControlModifier, Qt.ShiftModifier, Qt.ControlModifier | Qt.ShiftModifier]:
            if event.modifiers() == Qt.ControlModifier:
                new_pos.setX(self.startDragpos.x() + delta_x / 2)
                new_pos.setY(self.startDragpos.y() + delta_y / 2)
            elif event.modifiers() == Qt.ShiftModifier:
                new_pos.setX(self.startDragpos.x() + delta_x / 4)
                new_pos.setY(self.startDragpos.y() + delta_y / 4)
            elif event.modifiers() == Qt.ControlModifier | Qt.ShiftModifier:
                new_pos.setX(self.startDragpos.x() + delta_x / 8)
                new_pos.setY(self.startDragpos.y() + delta_y / 8)
            mouse_event = QMouseEvent(event.type(), new_pos, event.button(), event.buttons(), event.modifiers())
            super(Slider, self).mouseMoveEvent(mouse_event)
            self.setValue(self.value() - self._delta_value)
        else:
            super(Slider, self).mouseMoveEvent(event)

    def keyPressEvent(self, event):
        p = self.mapFromGlobal(QCursor.pos())
        self._start_drag_pos = p
        self._real_start_drag_pos = p
        self._default_value = 0
        super(Slider, self).keyPressEvent(event)

    def wheelEvent(self, event):
        if not self.hasFocus():
            event.ignore()
        else:
            super(Slider, self).wheelEvent(event)

    @property
    def slider_range(self):
        return self._slider_range


class DoubleSlider(Slider, object):
    doubleValueChanged = Signal(float)

    def __init__(self, parent=None, slider_range=None, default_value=0.0, maximum_value=1000000, dragger_steps=None):
        if slider_range is None:
            slider_range = (-100.0, 100.0)
        if dragger_steps is None:
            dragger_steps = FLOAT_SLIDER_DRAG_STEPS
        super(DoubleSlider, self).__init__(parent=parent, dragger_steps=dragger_steps, slider_range=slider_range)

        self._is_float = True
        self._maximum_value = abs(maximum_value)

        self.setOrientation(Qt.Horizontal)
        self.setMinimum(0)
        self.setMaximum(self._maximum_value)

        self.valueChanged.connect(self._on_value_changed)
        self.valueIncremented.connect(self._on_value_incremented)
        self.set_mapped_value(default_value, True)

    def mapped_value(self):
        return self._map_value(self.value())

    def set_mapped_value(self, value, block_signals=False):
        internal_value = self._unmap_value(value)
        if block_signals:
            self.blockSignals(True)
        self.setValue(internal_value)
        if self.signalsBlocked() and block_signals:
            self.blockSignals(False)

    def _map_value(self, in_value):
        """
        Internal function that converts slider integer value to slider float range value
        :param in_value:
        :return:
        """

        return mathlib.map_range_unclamped(
            in_value, self.minimum(), self.maximum(), self._slider_range[0], self._slider_range[1])

    def _unmap_value(self, out_value):
        """
        Converts mapped float value to slider integer value
        :param out_value:
        :return:
        """

        return int(mathlib.map_range_unclamped(
            out_value, self._slider_range[0], self._slider_range[1], self.minimum(), self.maximum()))

    def _on_value_changed(self, x):
        mapped_value = self._map_value(x)
        self.doubleValueChanged.emit(mapped_value)

    def _on_value_incremented(self, step):
        slider_internal_range = (self.minimum(), self.maximum())
        slider_dst = max(slider_internal_range) - min(slider_internal_range)
        value_dst = max(self._slider_range) - min(self._slider_range)
        factor = slider_dst / value_dst
        unmapped_step = step * factor
        current_internal_value = self.value()
        new_unmapped_value = current_internal_value + unmapped_step
        self.setValue(new_unmapped_value)


class HoudiniInputDragger(QWidget, object):
    """
    Widget that allow to drag values when mid click over widget.
    Right Drag increments values and Left Drag decreases value
    """

    def __init__(self, parent, factor, main_color=None, *args, **kwargs):
        super(HoudiniInputDragger, self).__init__(*args, **kwargs)

        self._parent = parent
        self._factor = factor
        self._main_color = main_color if main_color else QColor(215, 128, 26).getRgb()
        self._size = 35

        self.setAttribute(Qt.WA_Hover)
        self.setStyleSheet(self._get_style_sheet())
        self.setMinimumHeight(self._size)
        self.setMinimumWidth(self._size)
        self.setMaximumHeight(self._size)
        self.setMaximumWidth(self._size)

        main_layout = layouts.VerticalLayout(margins=(0, 0, 0, 0))
        self.setLayout(main_layout)

        frame_layout = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))
        self._frame = QGroupBox()
        self._frame.setLayout(frame_layout)
        main_layout.addWidget(self._frame)

        self._label = label.BaseLabel('+' + str(factor), parent=self)
        font = self._label.font()
        font.setPointSize(7)
        self._label.setFont(font)
        self._label.setAlignment(Qt.AlignCenter)
        frame_layout.addWidget(self._label)

        self.installEventFilter(self)
        self._label.installEventFilter(self)

    def eventFilter(self, obj, event):
        if event.type() == QEvent.HoverEnter:
            self.setStyleSheet(self._get_style_sheet(hover_style=True))
            self._parent.active_drag = self
            for drag in self._parent.drags:
                if drag != self:
                    drag.setStyleSheet(self._get_style_sheet())
        if event.type() == QEvent.HoverLeave:
            if event.pos().y() > self.height() or event.pos().y() < 0:
                self.setStyleSheet(self._get_style_sheet())
        if event.type() == QEvent.MouseMove:
            self._parent.eventFilter(self, event)

        return False

    @property
    def factor(self):
        return self._factor

    def _get_style_sheet(self, hover_style=False):
        if hover_style:
            return """
            QGroupBox{
                border: 0.5 solid darkgrey;
                background : %s;
                color: white;
            }
            QLabel{
                background: transparent;
                border: 0 solid transparent;
                color: white;
            }
            """ % "rgba%s" % str(self._main_color)
        else:
            return """
            QGroupBox{
                border: 0.5 solid darkgrey;
                background : black;
                color: white;
            }
            QLabel{
                background: transparent;
                border: 0 solid transparent;
                color: white;
            }
            """


class DraggerSlider(QDoubleSpinBox, object):
    """
    Slider that holds Houdini style draggers to drag values when mid click over them.
    Middle click to display draggers to change value by adding different delta values
    """

    valueIncremented = Signal(object)

    def __init__(self, label_text='', slider_type='float', buttons=False, decimals=3, dragger_steps=None,
                 apply_style=True, main_color=None, *args, **kwargs):
        super(DraggerSlider, self).__init__(*args, **kwargs)

        self._label_text = label_text
        self._main_color = main_color if main_color else QColor(215, 128, 26).getRgb()
        self._dragger_steps = dragger_steps or FLOAT_SLIDER_DRAG_STEPS
        self._is_float = slider_type == 'float'
        self._draggers = None

        self.setFocusPolicy(Qt.StrongFocus)
        if not self._is_float:
            self.setDecimals(0)
            self.setRange(qtutils.FLOAT_RANGE_MIN, qtutils.FLOAT_RANGE_MAX)
        else:
            self.setDecimals(decimals)
            self.setRange(qtutils.INT_RANGE_MIN, qtutils.INT_RANGE_MAX)
        if not buttons:
            self.setButtonSymbols(QAbstractSpinBox.NoButtons)
        if apply_style:
            self._label_font = QFont('Serif', 10, QFont.Bold)
            self.setStyleSheet(self._get_style_sheet())
        else:
            self._label_font = self.lineEdit().font()

        self.lineEdit().installEventFilter(self)
        self.installEventFilter(self)

    def wheelEvent(self, event):
        if not self.hasFocus():
            event.ignore()
        else:
            super(DraggerSlider, self).wheelEvent(event)

    def update(self):
        self.setStyleSheet(self._get_style_sheet())

    def paintEvent(self, event):
        super(DraggerSlider, self).paintEvent(event)

        p = QPainter()
        p.begin(self)
        p.setPen(color.DARK_GRAY)
        p.setFont(self._label_font)
        p.drawText(self.rect(), Qt.AlignCenter, self._label_text)
        p.end()

    def eventFilter(self, obj, event):
        if event.type() == QEvent.MouseButtonPress:
            if event.button() == Qt.MiddleButton:
                if not self._draggers:
                    self._draggers = SliderDraggers(self, self._is_float, dragger_steps=self._dragger_steps)
                    self._draggers.increment.connect(self._on_value_incremented)
                self._draggers.show()
                if self._is_float:
                    self._draggers.move(
                        self.mapToGlobal(QPoint(event.pos().x() - 1, event.pos().y() - self._draggers.height() / 2)))
                else:
                    self._draggers.move(
                        self.mapToGlobal(QPoint(event.pos().x() - 1, event.pos().y() - self._draggers.height() + 15)))

        return False

    def _get_style_sheet(self):
        return """
        QWidget{
            border: 1.25 solid black;
        }
        QSlider::groove:horizontal,
            QSlider::sub-page:horizontal {
            background: %s;
        }
        QSlider::add-page:horizontal,
            QSlider::sub-page:horizontal:disabled {
            background: rgb(32, 32, 32);
        }
        QSlider::add-page:horizontal:disabled {
            background: grey;
        }
        QSlider::handle:horizontal {
            width: 1px;
         }
        """ % "rgba%s" % str(self._main_color)

    def _on_value_incremented(self, step):
        self.valueIncremented.emit(step)
        self.setValue(self.value() + step)


@theme.mixin
class HoudiniDoubleSlider(QWidget, object):
    """
    Slider that encapsulates a DoubleSlider and Houdini draggers linked together
    """

    valueChanged = Signal(object)

    def __init__(self, parent, slider_type='float', style=0, name=None, slider_range=None, default_value=0.0,
                 dragger_steps=None, main_color=None, *args):
        if slider_range is None:
            slider_range = (-100.0, 100.0)
        if dragger_steps is None:
            dragger_steps = FLOAT_SLIDER_DRAG_STEPS
        super(HoudiniDoubleSlider, self).__init__(parent=parent, *args)

        h = 20
        self._parent = parent
        self._type = slider_type
        self._value = 0.0
        self._label = None
        self._style_type = style

        theme = self.theme()
        if theme:
            theme_color = theme.accent_color
            if core_color.string_is_hex(theme_color):
                theme_color = core_color.hex_to_rgb(theme_color)
                main_color = QColor(*theme_color).getRgb()

        self._main_color = main_color or QColor(215, 128, 26).getRgb()

        self.setMaximumHeight(h)
        self.setMinimumHeight(h)

        self._main_layout = layouts.HorizontalLayout(margins=(10, 0, 0, 0))
        self.setLayout(self._main_layout)

        self._input = DraggerSlider(slider_type=slider_type)
        self._input.setButtonSymbols(QAbstractSpinBox.NoButtons)
        self._input.setRange(slider_range[0], slider_range[1])
        self._input.setContentsMargins(0, 0, 0, 0)
        self._input.setMinimumWidth(60 if self._type == 'float' else 40)
        self._input.setMaximumWidth(60 if self._type == 'float' else 40)
        self._input.setMinimumHeight(h)
        self._input.setMaximumHeight(h)
        self._input.valueIncremented.connect(self._on_increment_value)

        if self._type == 'float':
            self._slider = DoubleSlider(parent=self, default_value=default_value, slider_range=slider_range,
                                        dragger_steps=dragger_steps)
        else:
            self._slider = Slider(parent=self, slider_range=slider_range)
            self._slider.valueIncremented.connect(self._on_increment_value)
        self._slider.setContentsMargins(0, 0, 0, 0)
        self._slider.setMinimumHeight(h)
        self._slider.setMaximumHeight(h)
        self._slider.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

        if name:
            self._label = label.BaseLabel(name + '  ', parent=self)
            self._main_layout.addWidget(self._label)
        self._main_layout.addWidget(self._input)
        self._main_layout.addWidget(self._slider)

        style_sheet = self._get_style_sheet(self._style_type)
        if self._style_type == 0:
            self._main_layout.setSpacing(0)
        self._slider.setStyleSheet(style_sheet)

        self._slider.valueChanged.connect(self._on_slider_value_changed)
        self._input.valueChanged.connect(self._on_houdini_slider_value_changed)

    def update(self):
        style_sheet = self._get_style_sheet(self._style_type)
        if self._style_type == 0:
            self._main_layout.setSpacing(0)
        self._slider.setStyleSheet(style_sheet)

    @property
    def minimum(self):
        return self._input.minimum()

    @property
    def maximum(self):
        return self._input.maximum()

    @property
    def _value_range(self):
        return self.maximum - self.minimum

    def _get_value(self):
        return self.value()

    def _set_value(self, value):
        with qt_contexts.block_signals(self):
            self.set_value(value)

    intValue = Property(int, _get_value, _set_value, user=True)
    floatValue = Property(float, _get_value, _set_value, user=True)

    def value(self):
        self._value = self._input.value()
        if self._type == 'int':
            self._value = int(self._value)

        return self._value

    def set_value(self, value):
        self._input.setValue(value)
        self._value = self._input.value()
        self.valueChanged.emit(self.value())
        # self._on_houdini_slider_value_changed(0)

    def set_decimals(self, decimals):
        self._input.setDecimals(decimals)

    def set_single_step(self, step):
        self._input.setSingleStep(step)

    def hide_label(self):
        if self._label:
            self._label.hide()

    def show_label(self):
        if self._label:
            self._label.show()

    def hide_slider(self):
        self._slider.hide()

    def show_slider(self):
        self._slider.show()

    def set_range(self, minimum_value, maximum_value):
        self._input.setRange(minimum_value, maximum_value)

    def _on_increment_value(self, step):
        if step == 0.0:
            return
        old = self._input.value()
        new = old + step
        self._input.setValue(new)
        self.valueChanged.emit(new)

    def _on_slider_value_changed(self, value):
        out_value = mathlib.map_range_unclamped(
            value, self._slider.minimum(), self._slider.maximum(), self._input.minimum(), self._input.maximum())
        with qt_contexts.block_signals(self._input):
            self._input.setValue(out_value)
        self.valueChanged.emit(out_value)

    def _on_houdini_slider_value_changed(self, value):
        in_value = mathlib.map_range_unclamped(
            self._input.value(), self._input.minimum(), self._input.maximum(),
            self._slider.minimum(), self._slider.maximum())
        with qt_contexts.block_signals(self._slider):
            self._slider.setValue(int(in_value))
        self.valueChanged.emit(value)

    def _get_style_sheet(self, style_type):
        if style_type == 0:
            return """
            QWidget{
                border: 1.25 solid black;
            }
            QSlider::groove:horizontal,
                QSlider::sub-page:horizontal {
                background: %s;
            }
            QSlider::add-page:horizontal,
                QSlider::sub-page:horizontal:disabled {
                background: rgb(32, 32, 32);
            }
            QSlider::add-page:horizontal:disabled {
                background: grey;
            }
            QSlider::handle:horizontal {
                width: 1px;
             }
            """ % "rgba%s" % str(self._main_color)
        else:
            return """
            QSlider::groove:horizontal {
                border: 1px solid #bbb;
                background: white;
                height: 3px;
                border-radius: 2px;
            }
            QSlider::sub-page:horizontal {
                background: %s;
                border: 0px solid #777;
                height: 3px;
                border-radius: 2px;
            }
            QSlider::add-page:horizontal {
                background: #fff;
                border: 1px solid #777;
                height: 3px;
                border-radius: 2px;
            }
            QSlider::handle:horizontal {
                background: qlineargradient(x1:0, y1:0, x2:1, y2:1,
                    stop:0 #eee, stop:1 #ccc);
                border: 1px solid #777;
                width: 4px;
                margin-top: -8px;
                margin-bottom: -8px;
                border-radius: 2px;
                height : 10px;
            }
            QSlider::handle:horizontal:hover {
                background: qlineargradient(x1:0, y1:0, x2:1, y2:1,
                    stop:0 #fff, stop:1 #ddd);
                border: 1px solid #444;
                border-radius: 2px;
            }
            QSlider::sub-page:horizontal:disabled {
                background: #bbb;
                border-color: #999;
            }
            QSlider::add-page:horizontal:disabled {
                background: #eee;
                border-color: #999;
            }
            QSlider::handle:horizontal:disabled {
                background: #eee;
                border: 1px solid #aaa;
                border-radius: 2px;
                height : 10;
            }
            """ % "rgba%s" % str(self._main_color)



from functools import partial

import tpDcc
from tpDcc.libs.qt.core import mixin
from tpDcc.libs.qt.widgets import dividers, message, group


class MessageExample(tpDcc.Window, mixin.FieldMixin):
    def __init__(self):
        super(MessageExample, self).__init__(title='Example for Messages')

    def ui(self):
        super(MessageExample, self).ui()

        self.main_layout.addWidget(dividers.Divider('different type'))
        self.main_layout.addWidget(message.BaseMessage(text='Information message', parent=self).info())
        self.main_layout.addWidget(message.BaseMessage(text='Success message', parent=self).success())
        self.main_layout.addWidget(message.BaseMessage(text='Warning message', parent=self).warning())
        self.main_layout.addWidget(message.BaseMessage(text='Error message', parent=self).error())
        self.main_layout.addWidget(dividers.Divider('closable'))
        self.main_layout.addWidget(message.BaseMessage(text='Closable message', parent=self).closable())
        self.main_layout.addWidget(dividers.Divider('data bind'))

        data_bind_msg = message.BaseMessage(parent=self)
        data_bind_msg.set_closable(True)

        self.register_field('msg', '')
        self.register_field('msg_type', message.MessageTypes.INFO)
        self.bind('msg', data_bind_msg, 'text')
        self.bind('msg_type', data_bind_msg, 'theme_type')

        button_grp = group.PushButtonGroup()
        button_grp.set_button_list([
            {'text': 'error', 'clicked': partial(self._on_change_msg, 'password is wrong', message.MessageTypes.ERROR)},
            {'text': 'success', 'clicked': partial(self._on_change_msg, 'Login success', message.MessageTypes.SUCCESS)},
            {'text': 'no more error', 'clicked': partial(self._on_change_msg, '', message.MessageTypes.ERROR)},
        ])
        self.main_layout.addWidget(button_grp)
        self.main_layout.addWidget(data_bind_msg)
        self.main_layout.addStretch()

    def _on_change_msg(self, msg_text, msg_type):
        self.set_field('msg_type', msg_type)
        self.set_field('msg', msg_text)


def run():
    return MessageExample()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains form field widgets
"""

from __future__ import print_function, division, absolute_import

import re
import logging
import traceback
from functools import partial

from Qt.QtCore import Qt, Signal
from Qt.QtWidgets import QSizePolicy, QFrame, QMenu, QFileDialog, QTextEdit
from Qt.QtGui import QCursor, QColor, QPixmap, QIntValidator

from tpDcc.managers import resources
from tpDcc.libs.python import decorators, path as path_utils
from tpDcc.libs.resources.core import color as qt_color
from tpDcc.libs.qt.core import contexts as qt_contexts
from tpDcc.libs.qt.widgets import layouts, label, buttons, checkbox, lineedit, combobox, color, sliders, dividers
from tpDcc.libs.qt.widgets import group, message, icon

LOGGER = logging.getLogger('tpDcc-libs-qt')


class FieldWidget(QFrame, object):

    valueChanged = Signal()

    DEFAULT_LAYOUT = 'horizontal'

    def __init__(self, parent=None, data=None, form_widget=None):
        super(FieldWidget, self).__init__(parent)

        self._data = data or dict()
        self._widget = None
        self._default = None
        self._required = None
        self._error_label = None
        self._menu_button = None
        self._action_result = None
        self._form_widget = None
        self._collapsed = False
        if form_widget:
            self.set_form_widget(form_widget)

        self.setObjectName('fieldWidget')

        direction = self._data.get('layout', self.DEFAULT_LAYOUT)
        if direction == 'vertical':
            main_layout = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))
        else:
            main_layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        self.setLayout(main_layout)

        self._label = label.BaseLabel(parent=self)
        self._label.setObjectName('label')
        self._label.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        main_layout.addWidget(self._label)

        self._layout2 = layouts.HorizontalLayout()
        main_layout.addLayout(self._layout2)
        if direction == 'vertical':
            self._label.setAlignment(Qt.AlignLeft | Qt .AlignVCenter)
        else:
            self._label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)

        widget = self.create_widget()
        if widget:
            self.set_widget(widget)

    # =================================================================================================================
    # ABSTRACT FUNCTIONS
    # =================================================================================================================

    @decorators.abstractmethod
    def value(self):
        """
        Returns the current value fo the field widget
        :return: variant
        """

        raise NotImplementedError('value function of FieldWidget is not implemented!')

    @decorators.abstractmethod
    def set_items(self, items):
        """
        Sets the items for the field widget
        :param items: list(str)
        """

        raise NotImplementedError('set_items function of FieldWidget is not implemented!')

    @decorators.abstractmethod
    def set_placeholder(self, placeholder):
        """
        Sets the placeholder text to de displayed for the widget
        :param placeholder: str
        """

        raise NotImplementedError('set_placeholder function of FieldWidget is not implemented!')

    # =================================================================================================================
    # STATIC FUNCTIONS
    # =================================================================================================================

    @staticmethod
    def to_title(name):
        """
        Converts camel case strings to title strings
        :param name: str
        :return: str
        """

        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1 \2', name)
        return re.sub('([a-z0-9])([A-Z])', r'\1 \2', s1).title()

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def name(self):
        """
        Returns the name of the field widget
        :return: str
        """

        return self.data()['name']

    def default_data(self):
        """
        Default data used by the field schema
        :return: dict
        """

        return dict()

    def is_collapsed(self):
        """
        Returns whether or not the field widget is collapsed/visible
        :return: bool
        """

        return self._collapsed

    def set_collapsed(self, flag):
        """
        Sets whether or not the field widget is collapsed/visible
        :param flag: bool
        """

        self._collapsed = flag

    def is_default(self):
        """
        Returns whether the current value is the same as the default value or not
        :return: bool
        """

        return self.value() == self.default()

    def default(self):
        """
        Returns the default value for the field widget
        :return: variant
        """

        return self._default

    def set_default(self, default):
        """
        Sets teh default value for the field widget
        :param default: variant
        """

        self._default = default

    def is_required(self):
        """
        Returns whether current field is required for the field widget or not
        :return: bool
        """

        return bool(self._required)

    def set_required(self, flag):
        """
        Sets whether the field is required for the field widget or not
        :param flag: bool
        """

        self._required = flag
        self.setProperty('required', flag)
        # self.setStyleSheet(self.styleSheet())

    def set_value(self, value):
        """
        Sets the value of the field widget
        :param value: variant
        """

        self._on_emit_value_changed()

    def state(self):
        """
        Returns the current state of the data
        :return: dict
        """

        return {
            'name': self._data['name'],
            'value': self.value()
        }

    def data(self):
        """
        Returns the data for the widget
        :return: dict
        """

        return self._data

    def set_data(self, data):
        """
        Sets the current state of the field widget using a dictionary
        :param data: dict
        """

        state = data

        with qt_contexts.block_signals(self):
            items = state.get('items', None)
            if items is not None:
                self.set_items(items)

            value = state.get('value', None)
            default = state.get('default', None)

            if default is not None:
                self.set_default(default)
            elif value is not None:
                self.set_default(value)

            if value is not None or (value and value != self.value()):
                try:
                    self.set_value(value)
                except TypeError:
                    LOGGER.exception(str(traceback.format_exc()))

            enabled = state.get('enabled', None)
            if enabled is not None:
                self.setEnabled(enabled)
                self._label.setEnabled(enabled)
            hidden = state.get('hidden', None)
            if hidden is not None:
                self.setHidden(hidden)
            required = state.get('required', None)
            if required is not None:
                self.set_required(required)
            error = state.get('error', None)
            if error is not None:
                self.set_error(error)
            error_visible = data.get('errorVisible')
            if error_visible is not None:
                self.set_error_visible(error_visible)
            tooltip = state.get('toolTip', None)
            if tooltip is not None:
                self.setToolTip(tooltip)
                self.setStatusTip(tooltip)
            style = state.get('style', None)
            if style is not None:
                self.setStyleSheet(style)
            title = self.title() or ''
            self.set_text(title)
            lbl = state.get('label')
            if lbl is not None:
                text = lbl.get('name', None)
                if text is not None:
                    self.set_text(text)
                visible = lbl.get('visible', None)
                if visible is not None:
                    self.label().setVisible(visible)

            # Menu items
            actions = state.get('actions', None)
            if actions is not None:
                self._menu_button.setVisible(True)
            menu = state.get('menu', None)
            if menu is not None:
                text = menu.get('name')
                if text is not None:
                    self._menu_button.setText(text)
                visible = menu.get('visible', True)
                self._menu_button.setVisible(visible)
            self._data.update(data)
            self.refresh()

    def has_error(self):
        """
        Returns whether or not this field contains any error
        :return: bool
        """

        return bool(self.data().get('error'))

    def set_error(self, message):
        """
        Sets the error message to be displayed for the field widget
        :param message: str
        """

        self._data['error'] = message
        self.refresh_error()

    def set_error_visible(self, flag):
        """
        Sets wehether or not the error message should be displayed
        :param flag: bool
        """

        self._data['errorVisible'] = flag
        self.refresh_error()

    def refresh_error(self):
        """
        Refreshes the error message with the current data
        """

        error = self._data.get('error', None)
        if self.has_error() and self.data().get('errorVisible', False):
            self._error_label.text = error
            self._error_label.setHidden(False)
            self.setToolTip(error)
        else:
            self._error_label.text = ''
            self._error_label.setHidden(True)
            self.setToolTip(self.data().get('toolTip'))

        self.refresh()

    # =================================================================================================================
    # UI
    # =================================================================================================================

    def label(self):
        """
        Returns label widget
        :return: QLabel
        """

        return self._label

    def set_text(self, text):
        """
        Sets the label text for the field widget
        :param text: str
        """

        self._label.setText(text)

    def form_widget(self):
        """
        Returns the form widget that contains the field widget
        :return: FormWidget
        """

        return self._form_widget

    def set_form_widget(self, form_widget):
        """
        Sets the form widget that contains the field widget
        :param form_widget: FormWidget
        """

        self._form_widget = form_widget

    def create_widget(self):
        """
        Creates the widget to be used by the field
        :return: QWidget or None
        """

        return None

    def title(self):
        """
        Returns the title to be displayed for the field
        :return: str
        """

        data = self.data()
        title = data.get('title', '') or data.get('name', '')
        if title:
            title = self.to_title(title)

        if self.is_required():
            title += '*'

        return title

    def widget(self):
        """
        Returns the widget used to set and get the field value
        :return: QWidget
        """

        return self._widget

    def set_widget(self, widget):
        """
        Sets the widget used to set and get the field value
        :param widget: QWidget
        """

        widget_layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))

        self._widget = widget
        self._widget.setParent(self)
        self._widget.setObjectName('widget')
        # self._widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

        self._menu_button = buttons.BaseButton()
        self._menu_button.setIcon(resources.icon('menu_dots'))
        self._menu_button.setHidden(True)
        self._menu_button.setObjectName('menuButton')
        # self._menu_button.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        self._menu_button.clicked.connect(self._on_menu_callback)

        widget_layout.addWidget(self._widget)
        widget_layout.addWidget(self._menu_button)

        layout = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))

        self._error_label = message.BaseMessage(parent=self).error().set_show_icon(False)
        self._error_label.setMaximumHeight(40)
        self._error_label.setHidden(True)
        self._error_label.setObjectName('errorLabel')
        self._error_label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

        layout.addLayout(widget_layout)
        layout.addWidget(self._error_label)

        self._layout2.addLayout(layout)

    def reset(self):
        """
        Resets the field widget back to its default values
        """

        self.set_state(self._data)

    def refresh(self):
        """
        Refresh the style properties of the field
        """

        direction = self._data.get('layout', self.DEFAULT_LAYOUT)
        self.setProperty('layout', direction)
        self.setProperty('default', self.is_default())
        if self.data().get('errorVisible', False):
            self.setProperty('error', self.has_error())

        self.setStyleSheet(self.styleSheet())

    # =================================================================================================================
    # INTERNAL
    # =================================================================================================================

    def _action_callback(self, callback):
        """
        Internal function that wraps schema callback to get the return value
        :param callback: fn
        """

        self._action_result = callback()

    # =================================================================================================================
    # CALLBACKS
    # =================================================================================================================

    def _on_menu_callback(self):
        """
        Internal callback function that is triggered when the menu button is clicked
        """

        callback = self.data().get('menu', {}).get('callback', self._on_show_menu)
        callback()

    def _on_show_menu(self):
        """
        Internal callback that shows field menu using the actions from the data
        """

        menu = QMenu(self)

        actions = self.data().get('actions', list())
        for action in actions:
            name = action.get('name', 'No name found')
            enabled = action.get('enabled', True)
            callback = action.get('callback')
            fn = partial(self._action_callback, callback)
            action = menu.addAction(name)
            action.setEnabled(enabled)
            action.triggered.connect(fn)

        point = QCursor.pos()
        point.setX(point.x() + 3)
        point.setY(point.y() + 3)

        self._action_result = None

        menu.exec_(point)

        if self._action_result is not None:
            self.set_value(self._action_result)

    def _on_emit_value_changed(self, *args):
        """
        Emits the value changed signal
        :param args: list
        """

        self.valueChanged.emit()
        self.refresh()


class IntFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(IntFieldWidget, self).__init__(*args, **kwargs)

        widget = sliders.BaseSlider(parent=self)
        widget.setMinimum(-50000000)
        widget.setMaximum(50000000)
        widget.valueChanged.connect(self._on_emit_value_changed)
        self.set_widget(widget)

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the widget
        :return: str
        """

        value = self.widget().value()

        return int(value)

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the widget
        :param value: str
        """

        if value == '':
            value = self.default()

        self.widget().setValue(int(value))


class BoolFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(BoolFieldWidget, self).__init__(*args, **kwargs)

        widget = checkbox.BaseCheckBox(parent=self)
        widget.stateChanged.connect(self._on_emit_value_changed)
        self.set_widget(widget)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def set_text(self, text):
        """
        Sets the label text for the field widget
        Override to support inline key.
        :param text: str
        """

        inline = self.data().get('inline')
        if inline:
            self.label().setText('')
            self.widget().setText(text)
        else:
            super(BoolFieldWidget, self).set_text(text)

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the checkbox
        :return: str
        """

        return bool(self.widget().isChecked())

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the checkbox
        :param value: str
        """

        self.widget().setChecked(value)
        super(BoolFieldWidget, self).set_value(value)


class EnumFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(EnumFieldWidget, self).__init__(*args, **kwargs)

        widget = combobox.BaseComboBox(parent=self)
        widget.currentIndexChanged.connect(self._on_emit_value_changed)
        self.set_widget(widget)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the combo box
        :return: str
        """

        return str(self.widget().currentText())

    def set_value(self, item):
        """
        Overrides FileWidget set_value function
        Sets the value of the combo box
        :param item: str
        """

        self.set_current_text(item)

    def set_items(self, items):
        """
        Overrides FieldWidget set_items function
        Sets the current items of the combo box
        :param items: list(str)
        """

        self.widget().clear()
        self.widget().addItems(items)

    def set_state(self, state):
        """
        Sets the current state with support for editable
        :param state: dict
        """

        super(EnumFieldWidget, self).set_state(state)
        editable = state.get('editable')
        if editable is not None:
            self.widget().setEditable(editable)

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def set_current_text(self, text):
        """
        Sets current text
        :param text: str
        """

        index = self.widget().findText(text, Qt.MatchExactly)
        if index != -1:
            self.widget().setCurrentIndex(index)
        else:
            LOGGER.warning('Cannot set the value for field {}'.format(self.name()))


class StringFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(StringFieldWidget, self).__init__(*args, **kwargs)

        widget = lineedit.BaseLineEdit(parent=self)
        widget.textChanged.connect(self._on_emit_value_changed)
        self.set_widget(widget)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the widget
        :return: str
        """

        return str(self.widget().text())

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the widget
        :param value: str
        """

        pos = self.widget().cursorPosition()

        with qt_contexts.block_signals(self.widget()):
            self.widget().setText(value)

        self.widget().setCursorPosition(pos)

        super(StringFieldWidget, self).set_value(value)


class StringDoubleFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(StringDoubleFieldWidget, self).__init__(*args, **kwargs)

        widget = QFrame(self)
        layout = layouts.HorizontalLayout(spacing=4, margins=(0, 0, 0, 0))
        widget.setLayout(layout)

        self._widget1 = lineedit.BaseLineEdit(parent=self)
        self._widget2 = lineedit.BaseLineEdit(parent=self)
        self._widget1.textChanged.connect(self._on_emit_value_changed)
        self._widget2.textChanged.connect(self._on_emit_value_changed)
        layout.addWidget(self._widget1)
        layout.addWidget(self._widget2)

        self.set_widget(widget)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def set_placeholder(self, placeholder):
        """
        Sets the placeholder text to de displayed for the widget
        :param placeholder: str
        """

        if isinstance(placeholder, (list, tuple)):
            if len(placeholder) >= 1:
                self._widget1.setPlaceholderText(placeholder[0])
            if len(placeholder) >= 2:
                self._widget2.setPlaceholderText(placeholder[1])
        else:
            self._widget1.setPlaceholderText(str(placeholder))

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the widget
        :return: str
        """

        value1 = self._widget1.text() or ''
        value2 = self._widget2.text() or ''

        return value1, value2

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the widget
        :param value: str
        """

        self._widget1.setText(str(value[0]))
        self._widget2.setText(str(value[1]))

        super(StringDoubleFieldWidget, self).set_value(value)


class PasswordFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(PasswordFieldWidget, self).__init__(*args, **kwargs)

        widget = label.QLineEdit(self)
        widget.setEchoMode(lineedit.BaseLineEdit.EchoMode.Password)
        widget.textChanged.connect(self._on_emit_value_changed)
        self.set_widget(widget)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the widget
        :return: str
        """

        return str(self.widget().text())

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the widget
        :param value: str
        """

        self.widget().setText(value)
        super(PasswordFieldWidget, self).set_value(value)


class TextFieldWidget(FieldWidget, object):

    DEFAULT_LAYOUT = 'vertical'

    def __init__(self, *args, **kwargs):
        super(TextFieldWidget, self).__init__(*args, **kwargs)

        widget = QTextEdit(parent=self)
        # widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        widget.textChanged.connect(self._on_emit_value_changed)
        self.set_widget(widget)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the text edit
        :return: str
        """

        return str(self.widget().toPlainText())

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the text edit
        :param value: str
        """

        print(value)
        self.widget().setText(value)
        super(TextFieldWidget, self).set_value(value)


class PathFieldWidget(StringFieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(PathFieldWidget, self).__init__(*args, **kwargs)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def set_data(self, data):
        """
        Overrides StringFieldWidget set_data function
        Adds a browse button to folder button
        :param data: dict
        """

        if 'menu' not in data:
            data['menu'] = {
                'callback': self._on_browse
            }

        super(PathFieldWidget, self).set_data(data)

    # =================================================================================================================
    # CALLBACKS
    # =================================================================================================================

    def _on_browse(self):
        """
        Opens file dialog
        """

        path = self.value()
        path = QFileDialog.getExistingDirectory(None, 'Browse Folder', path)
        if path:
            self.set_value(path)


class SeparatorFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(SeparatorFieldWidget, self).__init__(*args, **kwargs)

        widget = dividers.Divider()
        widget.setObjectName('widget')
        widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.set_widget(widget)
        if not self.data().get('title'):
            self.label().hide()

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the separator
        :return: str
        """

        return str(self.widget().get_text())

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the separator
        :param value: str
        """

        self.widget().set_text(str(value))


class LabelFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(LabelFieldWidget, self).__init__(*args, **kwargs)

        widget = label.RightElidedLabel(parent=self)
        widget.setAlignment(Qt.AlignVCenter)
        widget.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.set_widget(widget)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the label
        :return: str
        """

        try:
            return str(self.widget().text())
        except Exception:
            return str(path_utils.normalize_path(self.widget().text()))

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the label
        :param value: str
        """

        self.widget().setText(value)
        super(LabelFieldWidget, self).set_value(value)


class ImageFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(ImageFieldWidget, self).__init__(*args, **kwargs)

        self._value = ''
        self._pixmap = None

        widget = label.BaseLabel(parent=self)
        widget.setObjectName('widget')
        self.setStyleSheet('min-height: 32px;')
        widget.setScaledContents(False)
        widget.setAlignment(Qt.AlignHCenter)
        self.set_widget(widget)
        self.layout().addStretch()

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def resizeEvent(self, event):
        """
        Overrides FieldWidget resizeEvent function
        Called when teh field widget is resized
        :param event: QResizeEvent
        """

        self.update()

    def value(self):
        """
        Implements FieldWidget value function
        Returns the path of the image in disk
        :return: str
        """

        return self._value

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the path of the image in disk
        :param value: str
        """

        self._value = value
        self._pixmap = QPixmap(value)
        self.update()

    def update(self):
        """
        Updates the image depending on the size
        """

        if not self._pixmap:
            return

        width = self.widget().height()
        if self.widget().width() > self.widget().height():
            pixmap = self._pixmap.scaledToWidth(width, Qt.SmoothTransformation)
        else:
            pixmap = self._pixmap.scaledToHeight(width, Qt.SmoothTransformation)
        self.widget().setPixmap(pixmap)
        self.widget().setAlignment(Qt.AlignLeft)


class RadioFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(RadioFieldWidget, self).__init__(*args, **kwargs)

        self._radio_buttons = list()

        layout = layouts.VerticalLayout(margins=(0, 0, 0, 0))
        self._radio_frame = QFrame(self)
        self._radio_frame.setLayout(layout)

        self.set_widget(self._radio_frame)

        self.label().setStyleSheet('margin-top: 2px;')
        self.label().setAlignment(Qt.AlignRight | Qt.AlignTop)
        self.widget().setStyleSheet('margin-top: 2px;')

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def value(self):
        """
        Implements FieldWidget value function
        Returns the path of the image in disk
        :return: str
        """

        for radio_button in self._radio_buttons:
            if radio_button.isChecked():
                return radio_button.text()

        return ''

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the path of the image in disk
        :param value: str
        """

        for radio_button in self._radio_buttons:
            checked = radio_button.text() == value
            radio_button.setChecked(checked)

    def set_items(self, items):
        """
        Sets the items for the field widget
        :param items: list(str)
        """

        self.clear()

        for item in items:
            widget = buttons.BaseRadioButton(parent=self)
            widget.setText(item)
            widget.clicked.connect(self._on_emit_value_changed)
            self._radio_buttons.append(widget)
            self._radio_frame.layout().addWidget(widget)

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def clear(self):
        """
        Destroy all radio buttons
        """

        for radio_button in self._radio_buttons:
            radio_button.destroy()
            radio_button.close()
            radio_button.hide()


class SliderFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(SliderFieldWidget, self).__init__(*args, **kwargs)

        widget = sliders.DoubleSlider(parent=self)
        widget.setOrientation(Qt.Horizontal)
        widget.setObjectName('widget')
        widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        widget.valueChanged.connect(self._on_emit_value_changed)
        self.set_widget(widget)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the slider
        :return: str
        """

        return self.widget().value()

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the slider
        :param value: str
        """

        self.widget().setValue(value)


class RangeFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(RangeFieldWidget, self).__init__(*args, **kwargs)

        widget = QFrame(self)
        layout = layouts.HorizontalLayout(spacing=4, margins=(0, 0, 0, 0))
        widget.setLayout(layout)

        validator = QIntValidator(-50000000, 50000000, self)

        self._min_widget = lineedit.BaseLineEdit(parent=self)
        self._min_widget.setValidator(validator)
        self._min_widget.textChanged.connect(self._on_emit_value_changed)
        widget.layout().addWidget(self._min_widget)

        self._max_widget = lineedit.BaseLineEdit(parent=self)
        self._max_widget.setValidator(validator)
        self._max_widget.textChanged.connect(self._on_emit_value_changed)
        widget.layout().addWidget(self._max_widget)

        self.set_widget(widget)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def value(self):
        """
        Implements FieldWidget value function
        Returns the current range
        :return: list(int)
        """

        min_value = int(float(self._min_widget.text() or '0'))
        max_value = int(float(self._max_widget.text() or '0'))

        return min_value, max_value

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the current range
        :param value: list(int)
        """

        min_value, max_value = int(value[0], int(value[1]))
        self._min_widget.setText(str(min_value))
        self._max_widget.setText(str(max_value))

        super(RangeFieldWidget, self).set_value(value)


class ColorFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(ColorFieldWidget, self).__init__(*args, **kwargs)

        widget = color.ColorPicker()
        widget.setObjectName('widget')
        widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        widget.colorChanged.connect(self._on_color_changed)
        self.set_widget(widget)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def set_data(self, data):
        """
        Overrides FieldWidget data function
        :param data: dict
        """

        colors = data.get('colors')
        if colors:
            self.widget().set_colors(colors)

        super(ColorFieldWidget, self).set_data(data)

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the color picker
        :return: Color
        """

        return self.widget().current_color()

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the color picker
        :param value: str
        """

        self.widget().set_current_color(value)

    def set_items(self, items):
        """
        Sets the items for the field widget
        :param items: list(str)
        """

        self.widget().set_colors(items)

    # =================================================================================================================
    # CALLBACKS
    # =================================================================================================================

    def _on_color_changed(self, new_color):
        """
        Internal callback function triggered when the color changes from the color browser
        :param new_color: QColor
        """

        if isinstance(new_color, QColor):
            new_color = qt_color.Color(new_color).to_string()

        self.set_value(new_color)
        self._on_emit_value_changed()


class IconPickerFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(IconPickerFieldWidget, self).__init__(*args, **kwargs)

        self._value = 'rgb(100, 100, 100)'

        widget = icon.IconPicker()
        widget.setObjectName('widget')
        widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        widget.iconChanged.connect(self._on_icon_changed)
        self.set_widget(widget)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def set_data(self, data):
        """
        Overrides FieldWidget data function
        :param data: dict
        """

        colors = data.get('colors')
        if colors:
            self.widget().set_colors(colors)

        super(IconPickerFieldWidget, self).set_data(data)

    def value(self):
        """
        Overrides FieldWidget value function
        Returns the value of the color picker
        :return: Color
        """

        return self.widget().current_icon()

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the color picker
        :param value: str
        """

        self.widget().set_current_icon(value)

    def set_items(self, items):
        """
        Sets the items for the field widget
        :param items: list(str)
        """

        self.widget().set_icons(items)
        self.widget().menu_button().setVisible(False)

    # =================================================================================================================
    # CALLBACKS
    # =================================================================================================================

    def _on_icon_changed(self, icon):
        """
        Internal callback function that is triggered when the color changes from the color browser
        :param icon: QIcon
        """

        self.set_value(icon)
        self._on_emit_value_changed()


class GroupFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(GroupFieldWidget, self).__init__(*args, **kwargs)

        widget = group.GroupBoxWidget(self.data().get('title'), None)
        widget.set_checked(True)
        widget.toggled.connect(self.set_value)

        self.set_widget(widget)

        self.label().hide()

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def default_data(self):
        """
        Returns the default data for the group field
        :return: dict
        """

        return {
            'value': True,
            'persistent': True,
            'validate': False,
            'persistentKey': 'BaseItem'
        }

    def value(self):
        """
        Overrides FieldWidget value function
        Called when teh field widget is resized
        """

        return self.widget().is_checked()

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the path of the image in disk
        :param value: str
        """

        self.widget().set_checked(value)

        is_child = False
        if self.form_widget():
            for field_widget in self.form_widget().field_widgets():
                if field_widget.name() == self.name():
                    is_child = True
                    continue
                if is_child:
                    if isinstance(field_widget, GroupFieldWidget):
                        break
                    field_widget.set_collapsed(not value)
                    if value and field_widget.data().get('visible') is not None:
                        field_widget.setVisible(field_widget.data().get('visible'))
                    else:
                        field_widget.setVisible(value)


class ButtonGroupFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(ButtonGroupFieldWidget, self).__init__(*args, **kwargs)

        self._value = ''
        self._buttons = dict()

        items = self.data().get('items')

        widget = QFrame(self)
        layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        widget.setLayout(layout)

        index = 0
        for item in items:
            index += 1
            button = buttons.BaseButton(item, parent=self)
            button.setCheckable(True)
            button_callback = partial(self.set_value, item)
            button.clicked.connect(button_callback)
            self._buttons[item] = button
            if index == 1:
                button.setProperty('first', True)
            if index == len(items):
                button.setProperty('last', True)
            layout.addWidget(button)

        self.set_widget(widget)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def value(self):
        """
        Overrides FieldWidget value function
        """

        return self._value

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the path of the image in disk
        :param value: str
        """

        self._value = value

        with qt_contexts.block_signals(self):
            for button in list(self._buttons.values()):
                button.setChecked(False)
            if value in self._buttons:
                self._buttons[value].setChecked(True)

        super(ButtonGroupFieldWidget, self).set_value(value)

    def set_items(self, items):
        """
        Sets the items for the field widget
        Override to avoid not implemented error
        :param items: list(str)
        """

        pass


class TagsFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(TagsFieldWidget, self).__init__(*args, **kwargs)

        widget = combobox.BaseComboBox(parent=self)
        widget.setEditable(True)
        widget.editTextChanged.connect(self._on_emit_value_changed)
        widget.editTextChanged.connect(self._on_emit_value_changed)
        self.set_widget(widget)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def value(self):
        """
        Overrides FieldWidget value function
        Called when teh field widget is resized
        """

        try:
            current_text = self.widget().currentText()
        except AttributeError:
            current_text = self.widget().itemText(self.widget().currentIndex())

        return self._string_to_list(current_text)

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the path of the image in disk
        :param value: str
        """

        text = self._list_to_string(value)
        self.widget().setEditText(text)

    def set_items(self, items):
        """
        Sets the items for the field widget
        :param items: list(str)
        """

        self.widget().clear()
        self.widget().addItems(items or list())

    # =================================================================================================================
    # INTERNAL
    # =================================================================================================================

    def _list_to_string(self, data):
        """
        Internal function that returns a string from the given list
        ['a', 'b', c'] to 'a,b,c
        :param data: list
        :return: str
        """

        data = [str(item) for item in data]
        data = str(data).replace("[", "").replace("]", "")
        data = data.replace("'", "").replace('"', "")

        return data

    def _string_to_list(self, data):
        """
        Returns a list from the given string
        a,b,c to ['a', 'b', c']
        :param data: str
        :return: list(str)
        """

        data = '["' + str(data) + '"]'
        data = data.replace(' ', '')
        data = data.replace(',', '","')

        return eval(data)


class ObjectsFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(ObjectsFieldWidget, self).__init__(*args, **kwargs)

        self._value = list()

        widget = label.RightElidedLabel(parent=self)
        widget.setAlignment(Qt.AlignVCenter)
        widget.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.set_widget(widget)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def value(self):
        """
        Overrides FieldWidget value function
        """

        return self._value

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        :param value: str
        """

        if value:
            count = len(value)
            plural = 's' if count > 1 else ''
            msg = '{} object{} selected for saving'.format(count, plural)
        else:
            msg = 'Nothing selected for saving'

        self._value = value
        self.widget().setText(msg)
        super(ObjectsFieldWidget, self).set_value(value)


FIELD_WIDGET_REGISTRY = {
    'int': IntFieldWidget,
    'bool': BoolFieldWidget,
    'enum': EnumFieldWidget,
    'string': StringFieldWidget,
    'stringDouble': StringDoubleFieldWidget,
    'password': PasswordFieldWidget,
    'text': TextFieldWidget,
    'path': PathFieldWidget,
    'radio': RadioFieldWidget,
    'separator': SeparatorFieldWidget,
    'label': LabelFieldWidget,
    'image': ImageFieldWidget,
    'slider': SliderFieldWidget,
    'range': RangeFieldWidget,
    'color': ColorFieldWidget,
    'iconPicker': IconPickerFieldWidget,
    'group': GroupFieldWidget,
    'buttonGroup': ButtonGroupFieldWidget,
    'tags': TagsFieldWidget,
    'objects': ObjectsFieldWidget,
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains manager to handle tools
"""

from __future__ import print_function, division, absolute_import

import os
import logging
import weakref
import webbrowser
from functools import partial

from Qt.QtCore import Qt, Signal, QSize, QEvent
from Qt.QtWidgets import QApplication, QSizePolicy, QWidget, QFrame, QPlainTextEdit, QDialog
from Qt.QtWidgets import QMenu, QAction
from Qt.QtGui import QCursor, QPixmap, QFont

from tpDcc import dcc
from tpDcc.core import dcc as core_dcc
from tpDcc.managers import configs, resources
from tpDcc.libs.python import python, osplatform, process, color, win32
from tpDcc.libs.qt.core import qtutils, base, preferences
from tpDcc.libs.qt.widgets import layouts, label, stack, buttons, switch, gif, dividers, theme

LOGGER = logging.getLogger('tpDcc-libs-qt')


class ToolsetPropertiesDict(python.ObjectDict, object):

    def update(self, others, convert_dict=False):
        """
        Updates dicts found in the item as well
        :param others: dict, dictionary to update this object with
        :param convert_dict: bool, Convert the dictionaries in others items to ToolsetPropertyDict
        """

        if convert_dict:
            for k, d in others.items():
                others[k] = ToolsetPropertiesDict(**d) if type(d) == dict else others[k]

        super(ToolsetPropertiesDict, self).update(others)


class ToolsetWidget(stack.StackItem, object):

    ID = 'toolsetID'
    CONFIG = None
    EXTENSION = 'toolset'
    ATTACHER_CLASS = None

    StartLargest = -1
    StartSmallest = 0

    displaySwitched = Signal()
    updatedPropertyRequested = Signal()
    savePropertyRequested = Signal()
    windowClosed = Signal()
    toolsetClosed = Signal()
    toolsetHidden = Signal()
    toolsetShown = Signal()
    toolsetDragged = Signal()
    toolsetDropped = Signal()
    toolsetDragCancelled = Signal()
    toolsetActivated = Signal()
    toolsetDeactivated = Signal()
    helpModeChanged = Signal(bool)

    def __init__(self, tree_widget=None, icon_color=(255, 255, 255), parent=None, *args, **kwargs):

        self._block_save = False
        self._show_warnings = True
        self._icon = self.CONFIG.get('icon') if self.CONFIG else None
        self._icon_color = icon_color
        self._tree_widget = tree_widget
        self._stacked_widget = None
        self._prev_stack_index = None
        self._display_mode_button = None
        self._manual_button = None
        self._help_button = None
        self._help_widget = None
        self._settings_button = None
        self._connect_button = None
        self._widgets = list()
        self._attacher = None
        self._client = None
        self._help_mode = False
        self._help_event = None
        self._settings = kwargs.get('settings', None)
        self._dev = kwargs.get('dev', False)
        self._properties = self._setup_properties()
        self._dcc_actions = list()
        title = self.CONFIG.data.get('label', '') if self.CONFIG else ''
        collapsable = kwargs.get('collapsable', True)
        show_item_icon = kwargs.get('show_item_icon', True)

        super(ToolsetWidget, self).__init__(
            title=title, collapsed=True, icon=self._icon, shift_arrows_enabled=False, title_editable=False,
            parent=parent or tree_widget, collapsable=collapsable, show_item_icon=show_item_icon,
            delete_button_enabled=True if tree_widget else False,
        )

    # =================================================================================================================
    # PROPERTIES
    # =================================================================================================================

    @property
    def stacked_widget(self):
        return self._stacked_widget

    @property
    def attacher(self):
        if not self._attacher:
            return
        return self._attacher()

    @property
    def dev(self):
        return self._dev

    @property
    def settings(self):
        return self._settings

    @property
    def properties(self):
        return self._properties

    @property
    def info_mode(self):
        return self._help_mode

    @property
    def client(self):
        return self._client() if self._client else None

    # =================================================================================================================
    # TO OVERRIDE
    # =================================================================================================================

    def initialize_properties(self):
        """
        Initializes properties of the toolset widget
        :return: list
        """

        return list()

    def pre_content_setup(self):
        """
        Function that is called before toolset contents are created
        Override in specific toolset widgets
        """

        pass

    def contents(self):
        """
        Returns toolset widget contentS
        Override in specific toolset widgets
        :return:
        """

        pass

    def post_content_setup(self):
        """
        Function that is called after toolset contents are created
        Override in specific toolset widgets
        """

        pass

    def help_mode(self, flag):
        """
        Function that can be used to show custom tooltip widgets for current toolset
        Override in specific toolset widgets
        :param flag: bool, Whether or not help mode is enabled
        """

        pass

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def ui(self):
        super(ToolsetWidget, self).ui()

        self.setContentsMargins(0, 0, 0, 0)
        self._contents_layout.setContentsMargins(0, 0, 0, 0)
        self._contents_layout.setSpacing(0)

        self._stacked_widget = stack.SlidingOpacityStackedWidget(self)
        # NOTE: tpDcc style uses this objectName to apply specific style to this widget
        self._stacked_widget.setObjectName('toolsetStackedWidget')
        self._stacked_widget.setContentsMargins(0, 0, 0, 0)
        self._stacked_widget.setLineWidth(0)
        self.main_layout.addWidget(self._stacked_widget)

        self.show_expand_indicator(False)
        self.set_title_text_mouse_transparent(True)

        self._display_mode_button = DisplayModeButton(color=self._icon_color, size=16, parent=self)
        self._display_mode_button.setFixedSize(QSize(22, 22))
        connect_widget = QWidget()
        connect_layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        connect_widget.setLayout(connect_layout)
        self._connect_button = buttons.BaseToolButton(parent=self).image('connect').icon_only()
        self._connect_settings_button = buttons.BaseToolButton(parent=self).image('menu_dots').icon_only()
        self._connect_settings_button.setPopupMode(self._connect_settings_button.InstantPopup)
        self._connect_settings_button.setStyleSheet('QToolButton::menu-indicator{width:0px;}')
        self._connect_button.setFixedSize(QSize(22, 22))
        self._connect_button.setEnabled(False)
        self._connect_button.setToolTip('No connected to any DCC')
        self._dcc_button = buttons.BaseToolButton(parent=self).icon_only()
        connect_layout.addWidget(self._connect_button)
        connect_layout.addWidget(self._connect_settings_button)
        connect_layout.addWidget(self._dcc_button)
        self._manual_button = buttons.BaseMenuButton(parent=self)
        self._manual_button.set_icon(resources.icon('manual'))
        self._manual_button.setFixedSize(QSize(22, 22))
        self._help_button = buttons.BaseMenuButton(parent=self)
        self._help_button.set_icon(resources.icon('help'))
        self._help_button.setFixedSize(QSize(22, 22))
        self._help_switch = switch.SwitchWidget(parent=self)
        self._settings_button = buttons.BaseMenuButton(parent=self)
        self._settings_button.set_icon(resources.icon('settings'))
        self._settings_button.setFixedSize(QSize(22, 22))
        self._help_widget = ToolsetHelpWidget()

        empty_widget = QWidget()
        empty_layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        empty_widget.setLayout(empty_layout)
        empty_layout.addStretch()
        empty_label = label.BaseLabel('Tool has no UI')
        empty_label.theme_level = label.BaseLabel.Levels.H1
        empty_layout.addWidget(empty_label)
        empty_layout.addStretch()

        self._preferences_widget = preferences.PreferencesWidget(parent=self)

        # We call if after setting all buttons
        self.set_icon_color(self._icon_color)

        self.visual_update(collapse=True)

        self._dccs_menu = QMenu(self)

        display_button_pos = 7
        self._title_frame.horizontal_layout.insertWidget(display_button_pos - 1, self._manual_button)
        self._title_frame.horizontal_layout.insertWidget(display_button_pos - 1, self._help_switch)
        self._title_frame.horizontal_layout.insertWidget(display_button_pos - 1, self._help_button)
        self._title_frame.horizontal_layout.insertWidget(display_button_pos - 1, self._settings_button)
        self._title_frame.horizontal_layout.insertWidget(0, connect_widget)
        self._title_frame.horizontal_layout.insertWidget(display_button_pos, self._display_mode_button)
        self._title_frame.horizontal_layout.setSpacing(0)
        self._title_frame.horizontal_layout.setContentsMargins(0, 0, 0, 0)
        self._title_frame.delete_button.setIconSize(QSize(12, 12))
        self._title_frame.item_icon_button.setIconSize(QSize(20, 20))

        font = QFont()
        font.setBold(True)
        self.setFont(font)

        if not dcc.is_standalone():
            self._connect_button.setVisible(False)
            self._connect_settings_button.setVisible(False)
            self._dcc_button.setVisible(False)

        self._stacked_widget.addWidget(empty_widget)
        self._stacked_widget.addWidget(self._widget_hider)
        self._stacked_widget.addWidget(self._preferences_widget)

        self._widget_hider.setVisible(False)
        self._preferences_widget.setVisible(False)

    def setup_signals(self):
        super(ToolsetWidget, self).setup_signals()

        self._title_frame.mouseReleaseEvent = self._on_activate_event

        self._connect_button.clicked.connect(self._on_focus_dcc)
        self._display_mode_button.clicked.connect(self.set_current_index)
        self._display_mode_button.clicked.connect(lambda: self.displaySwitched.emit())
        self.displaySwitched.connect(lambda: self.updateRequested.emit())

        self._manual_button.leftClicked.connect(self._on_open_help)
        self._help_button.leftClicked.connect(self._on_toggle_info_switch)
        self._help_switch.toggled.connect(self._on_toggle_help_mode)
        self._settings_button.leftClicked.connect(self._on_show_preferences_dialog)
        self._preferences_widget.closed.connect(self._on_close_preferences_window)

        self.deletePressed.connect(self._on_stop_callbacks)

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def initialize(self, client=None):

        if client:
            client.dccDisconnected.connect(self._on_dcc_disconnected)
            status = client.get_status_level()
            status_message = client.get_status_message()
            self._reset_connect_button(status_message, status)
            self._client = weakref.ref(client)
            self._connect_button.setVisible(True)

        if not client or not dcc.is_standalone():
            self._connect_button.setVisible(False)
            self._connect_settings_button.setVisible(False)
            self._dcc_button.setVisible(False)
        else:
            self._setup_connect_settings_menu()
            self._refresh_dcc_actions()
            self._refresh_connect_settings_menu()
            self._refresh_dcc_button()

        self.pre_content_setup()
        toolset_contents = self.contents() or list()
        for toolset_widget in toolset_contents:
            self.add_stacked_widget(toolset_widget)
        self._stacked_widget.setCurrentIndex(1) if self.count() > 0 else self._stacked_widget.setCurrentIndex(0)
        self.post_content_setup()
        self.update_display_button()
        self.expand()

    def set_attacher(self, attacher):
        """
        Sets attacher of this toolset
        The attacher is the window this toolset belongs to. Can be a custom Qt window or Hub
        :param attacher:
        """

        self._attacher = weakref.ref(attacher)

        if hasattr(self.attacher, 'settings'):
            # Setup standard settings and attacher preferences
            self._preferences_widget.set_settings(self.attacher.settings())
            self._theme_prefs_widget = theme.ThemePreferenceWidget(
                theme=self.attacher.theme(), parent=self._preferences_widget)
            self._preferences_widget.add_category(self._theme_prefs_widget.CATEGORY, self._theme_prefs_widget)
            # self.setup_attacher_settings(attacher)

        self.attacher.closed.connect(self._on_attacher_closed)
        self.attacher.themeUpdated.connect(self.reload_theme)
        self.attacher.styleReloaded.connect(self.reload_theme)

        self.reload_theme(self.attacher.theme())

    def setup_attacher_settings(self, attacher):
        """
        Function that can be used to setup all the attacher settings widgets
        :return:
        """

        if not attacher:
            return

        if not hasattr(attacher, 'get_settings_widgets') or not callable(attacher.get_settings_widgets):
            return

        settings_widgets = attacher.get_settings_widgets()
        if not settings_widgets:
            return

    def get_icon(self):
        """
        Returns toolset icon
        :return: QIcon
        """

        if not self._title_frame:
            return resources.icon('tpdcc')

        return resources.icon(self._title_frame.item_icon or 'tpdcc')

    def widgets(self):
        """
        Returns all display widgets
        :return: list
        """

        return self._widgets

    def display_index(self):
        """
        Current index of the stacked widget
        :return: int
        """

        return self._stacked_widget.currentIndex()

    def current_widget(self):
        """
        Returns current active widget
        :return: QWidget
        """

        return self.widgets()[self.display_index()]

    def widget_at(self, index):
        """
        Returns stacked widget at given index
        :param index: int
        :return: QWidget
        """

        return self._stacked_widget.widget(index)

    def count(self):
        """
        Returns the total amount of widgets added to the contents stack
        :return: int
        """

        return len(self._widgets)

    def item_at(self, index):
        """
        Returns widget located in given index of the toolset stack
        :param index: int
        :return: QWidget or None
        """

        return self._stacked_widget.widget(index)

    def add_stacked_widget(self, widget):
        """
        Adds a new widget to the stack
        :param widget: QWidget
        """

        if not widget:
            raise ValueError(
                'Toolset "{}" contents() must return a list of widgets! None found.'.format(
                    str(self.__class__.__name__)))

        self._widgets.append(widget)
        widget.setVisible(False)
        widget.setProperty('color', self._icon_color)
        widget.setParent(self._widget_hider)
        self._contents_layout.addWidget(widget)

        self._widgets[0].setVisible(True)

    def visual_update(self, collapse=True):
        """
        Updates visual based on opened or closed state
        :param collapse: bool
        """

        if collapse:
            self.set_icon_color(color.desaturate(self._icon_color, 0.75), set_color=False)
            self.title_text_widget().setObjectName('disabled')
        else:
            self.set_icon_color(self._icon_color)
            self.title_text_widget().setObjectName('active')

        qtutils.update_widget_style(self.title_text_widget())

        self.setUpdatesEnabled(False)
        self.updatedPropertyRequested.emit()
        self.setUpdatesEnabled(True)

    def update_display_button(self):
        """
        Updates the display button based on number of widgets in the stack
        """

        self.set_displays(self.count())

    def set_displays(self, displays):
        """
        Sets displays
        :param displays: list
        """

        if displays in [ToolsetDisplays.Single, ToolsetDisplays.Double, ToolsetDisplays.Triple]:
            self._display_mode_button.set_displays(displays)
        else:
            self._display_mode_button.setVisible(False)

    def block_save(self, flag):
        """
        Sets whether data saving is enabled or not
        :param flag: bool
        """

        self._block_save = flag

    def set_active(self, active=True, emit=True):
        """
        Sets whether toolset widget is active or not
        :param active: bool
        :param emit: bool
        """

        if active:
            self.expand(emit)
        else:
            self.collapse(emit)
        self.visual_update(collapse=not active)

    def set_current_index(self, index):

        if not self._widgets:
            return

        self.block_save(True)

        for i in range(self.count()):
            w = self._widgets[i]
            w.setSizePolicy(w.sizePolicy().horizontalPolicy(), QSizePolicy.Ignored)
            w.setVisible(False)

        self._stacked_widget.setCurrentIndex(1)

        widget = self._widgets[index]
        if widget:
            widget.setSizePolicy(widget.sizePolicy().horizontalPolicy(), QSizePolicy.Expanding)
            widget.setVisible(True)
        else:
            LOGGER.warning('Widget not found!')

        self.block_save(False)

    def set_icon_color(self, color=(255, 255, 255), set_color=True):
        """
        Sets the icon color for all the icons
        :param color: tuple(int, int, int), color in RGB integer values in 0-255 range
        :param set_color: bool, saves current color overwriting cache variable
        """

        if set_color:
            self._icon_color = color
        darken = 0.8
        self.set_item_icon_color(color)
        self._display_mode_button.set_icon_color(color)
        self._help_button.set_icon_color((color[0] * darken, color[1] * darken, color[2] * darken))
        self._manual_button.set_icon_color((color[0] * darken, color[1] * darken, color[2] * darken))
        self._title_frame.delete_button.set_icon_color(color)

    def populate_widgets(self):

        property_widgets = self.property_widgets()

        # self.savePropertyRequested.connect(self.save_properties)
        # self.updatedPropertyRequested.connect(self.update_properties)

    # def linked_properties(self, widget):
    #     """
    #     Returns list of linkable properties from widgets children
    #     :param widget:  list
    #     :return:
    #     """
    #
    #     for attr in widget.__dict__:
    #         if type(getattr(widget, attr)) in
    #
    def auto_link_properties(self, widgets):
        """
        Add link properties if allowed
        :param widgets: list
        """

        if not self.CONFIG and self.CONFIG.get('auto_link_properties', True):
            return

        new_properties = list()
        names = list()
        # for link_property in self._linked

    def property_widgets(self, widget=None, current_widget=False):
        """
        Get properties widgets from the stack
        :param widget: QWidget
        :param current_widget: bool, Set True to only search the current active widget in the stack
        :return:
        """

        result = list()

        if current_widget:
            widget = widget or self._stacked_widget.current_widget()
        else:
            widget = widget or self._stacked_widget

        for child in qtutils.iterate_children(widget, skip='skipChildren'):
            if child.property('prop') is not None:
                result.append(child)

        return result

    def reload_theme(self, theme):
        if not theme:
            return
        stylesheet = theme.stylesheet()
        self.setStyleSheet(stylesheet)

        # cached_icons_keys = icon.IconCache._resources_keys_cache.copy()
        # keys_names_mapping = icon.IconCache._resources_names_keys_mapping
        #
        # print('gogogog')
        #
        # # Update icons taking into account the new theme
        # new_icons_to_apply = dict()
        # for icon_key, icon_object in cached_icons_keys.items():
        #     icon_full_name = keys_names_mapping.get(icon_key, None)
        #     if not icon_full_name:
        #         continue
        #     icon_name, icon_ext = os.path.splitext(icon_full_name)
        #     new_icon = resources.icon(icon_name, extension=icon_ext, theme=theme.name())
        #     if not new_icon or new_icon.isNull():
        #         continue
        #     new_icons_to_apply[icon_key] = new_icon
        #
        # # TODO: This can by quite hard to process. Find a better approach for this
        # all_widgets = self.findChildren(QWidget)
        # for w in all_widgets:
        #     if hasattr(w, 'setIcon'):
        #         try:
        #             curr_icon = w.icon()
        #             curr_icon_key = curr_icon.cacheKey()
        #             if curr_icon_key not in new_icons_to_apply:
        #                 continue
        #             w.setIcon(new_icons_to_apply[curr_icon_key])
        #         except Exception:
        #             pass

    # =================================================================================================================
    # INTERNAL
    # =================================================================================================================

    def _setup_properties(self, properties=None):

        tool_props = properties or self.initialize_properties()
        instance_props = ToolsetPropertiesDict()
        for prop in tool_props:
            instance_props[prop['name']] = ToolsetPropertiesDict(**prop)
            if 'default' not in instance_props[['name']]:
                instance_props[prop['name']].default = instance_props[prop['name']].value

        return instance_props

    def _stop_selection_callback(self):
        print('Stop Selection Callbacks ...')

    def _reset_connect_button(self, text='No connected to any DCC', severity='warning'):
        self._connect_button.setEnabled(False)
        self._connect_button.setToolTip(str(text))

        if severity == 'warning':
            LOGGER.warning(text)
            self._connect_button.setStyleSheet('background-color: #e4c019')
        elif severity == 'error':
            LOGGER.error(text)
            self._connect_button.setStyleSheet('background-color: #bc3030')
        else:
            LOGGER.info(text)
            self._connect_button.setStyleSheet('')
            self._connect_button.setEnabled(True)

    def _setup_connect_settings_menu(self):
        self._connect_settings_menu = QMenu(self)
        self._refresh_action = QAction(resources.icon('refresh'), 'Refresh', self)
        self._connect_settings_menu.addAction(self._refresh_action)
        self._connect_settings_menu.addSeparator()

        self._refresh_action.triggered.connect(self._on_refresh_client)

        self._connect_settings_button.setMenu(self._connect_settings_menu)

    def _refresh_dcc_actions(self):
        for action in self._dcc_actions:
            self._dcc_actions.remove(action)
            self._connect_settings_menu.removeAction(action)

        config_dict = configs.get_tool_config(self.ID) or dict() if self.ID else dict()
        supported_dccs = config_dict.get('supported_dccs', dict()) if config_dict else dict()
        if not supported_dccs:
            return

        for dcc_name, nice_name in core_dcc.Dccs.nice_names.items():
            if dcc_name not in supported_dccs:
                continue
            process_name = core_dcc.Dccs.executables.get(dcc_name, dict()).get(osplatform.get_platform(), None)
            if not process_name:
                continue
            process_running = process.check_if_process_is_running(process_name)
            if not process_running:
                continue
            dcc_icon = resources.icon(dcc_name)
            dcc_action = QAction(dcc_icon, nice_name, self)
            dcc_action.triggered.connect(partial(self._on_dcc_selected, dcc_name))
            self._connect_settings_menu.addAction(dcc_action)
            self._dcc_actions.append(dcc_action)

    def _refresh_connect_settings_menu(self):
        self._refresh_action.setVisible(not self.client.connected)

        if not self.client.connected:
            self._connect_settings_button.setVisible(True)
        else:
            if len(self._dcc_actions) > 1:
                self._connect_settings_button.setVisible(True)
            else:
                self._connect_settings_button.setVisible(False)

    def _refresh_dcc_button(self):
        self._dcc_button.setVisible(self.client.connected)

        if self.client.connected:
            dcc_icon = resources.icon(self.client.get_name())
            if dcc_icon.isNull():
                dcc_icon = resources.icon('tpDcc')
            self._dcc_button.setIcon(dcc_icon)
            self._dcc_button.setToolTip('{} - {}'.format(self.client.get_name(), self.client.get_version()))

    # =================================================================================================================
    # CALLBACKS
    # =================================================================================================================

    def _on_attacher_closed(self):
        """
        Internal callback function that is called before attacher is closed
        Useful to do custom things before closing a tool
        """

        pass

    def _on_activate_event(self, event, emit=True):
        self._on_toggle_contents(emit=emit)
        event.ignore()

    def _on_open_help(self):
        """
        Internal callback function that is called when help button is clicked
        """

        url = self.CONFIG.get('help_url', '') if self.CONFIG else ''
        if not url:
            return
        webbrowser.open(url)

    def _on_toggle_info_switch(self):
        self._help_switch.setChecked(not self._help_switch.isChecked())

    def _on_toggle_help_mode(self, flag):
        """
        Internal callback function that toggles info mode
        """

        self._help_mode = flag

        if not self._help_event:
            self._help_event = self.help_mode(self._help_mode)

        if self._help_event:
            if self._help_mode:
                self._help_event.help_mode = True
            else:
                self._help_event.help_mode = False
                self._help_event.current_tooltip.close()
                self._help_event.close()
                self._help_event = None

        self.helpModeChanged.emit(self._help_mode)

    def _on_stop_callbacks(self):
        self._stop_selection_callback()

    def _on_show_preferences_dialog(self):

        if not self.attacher:
            return

        if self._prev_stack_index is None:
            self._prev_stack_index = self._stacked_widget.currentIndex()
        self._stacked_widget.setCurrentIndex(2)

    def _on_close_preferences_window(self, *args):
        self._stacked_widget.setCurrentIndex(self._prev_stack_index)
        if self.attacher:
            self.reload_theme(self.attacher.theme())
        self._prev_stack_index = None

    def _on_dcc_disconnected(self):
        self._connect_button.setEnabled(False)
        self._connect_button.setToolTip('No connected to any DCC')

    def _on_focus_dcc(self):
        if not self.client:
            return

        _, _, dcc_pid = self.client.get_dcc_info()
        if not dcc_pid:
            return

        win32.focus_window_from_pid(dcc_pid)
        win32.focus_window_from_pid(os.getpid(), restore=False)

    def _on_refresh_client(self):
        self.client.connect()
        self.client.update_client(tool_id=self.ID)
        status = self.client.get_status_level()
        status_message = self.client.get_status_message()
        self._reset_connect_button(status_message, status)
        self._refresh_connect_settings_menu()
        self._refresh_dcc_button()

    def _on_dcc_selected(self, dcc_name):
        old_port = self.client._port
        dcc_port = core_dcc.dcc_port(self.client.PORT, dcc_name=dcc_name)
        if old_port == dcc_port:
            return

        try:
            self.client.disconnect()
        except Exception:
            pass
        valid = self.client.connect(dcc_port)
        if not valid:
            self.client.connect(old_port)

        self.client.update_client(tool_id=self.ID)
        status = self.client.get_status_level()
        status_message = self.client.get_status_message()
        self._reset_connect_button(status_message, status)
        self._refresh_connect_settings_menu()
        self._refresh_dcc_button()


class ToolsetDisplays(object):
    """
    Display modes for the toolset widget items
    """

    Single = 1
    Double = 2
    Triple = 3


class DisplayModeButton(buttons.BaseMenuButton, object):

    FIRST_INDEX = 1
    LAST_INDEX = -1

    clicked = Signal(object)

    def __init__(self, parent=None, size=16, color=(255, 255, 255), initial_index=FIRST_INDEX):
        super(DisplayModeButton, self).__init__(parent=parent)

        menu_icon_double_names = ['menu_double_empty', 'menu_double_one', 'menu_double_full']
        menu_icon_triple_names = ['menu_triple_empty', 'menu_triple_one', 'menu_triple_two', 'menu_triple_full']
        self._menu_icon_double = [resources.icon(menu_icon) for menu_icon in menu_icon_double_names]
        self._menu_icon_triple = [resources.icon(menu_icon) for menu_icon in menu_icon_triple_names]

        self._current_icon = None
        self._icons = None
        self._displays = None
        self._initial_display = initial_index
        self._current_size = size
        self._icon_color = color
        self._highlight_offset = 40

        self.setIconSize(QSize(size, size))
        self.set_displays(ToolsetDisplays.Double)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def mouseReleaseEvent(self, event):
        """
        Overrides base QWidget mouseReleaseEvent
        :param event: QMouseEvent
        """

        new_icon, new_index = self.next_icon()
        self.set_icon(new_icon, size=self._current_size, colors=self._icon_color)
        self._current_icon = new_icon
        self.clicked.emit(new_index - 1)

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def set_displays(self, displays=ToolsetDisplays.Triple):
        """
        Sets the number of displays
        :param displays:
        """

        self._displays = displays
        self.show()

        if displays == 3:
            self._icons = self._menu_icon_triple
        elif displays == 2:
            self._icons = self._menu_icon_double
        elif displays == 1:
            self.hide()
        else:
            LOGGER.error('only 2 or 3 displays are available!')

        self._current_icon = self._icons[self._initial_display]
        self.set_icon_index(self._initial_display)

    def set_icon_index(self, index, size=None, color=None):
        """
        Sets icon by its display index
        :param index: int
        :param size:
        :param color:
        """

        if size is None:
            size = self._current_size
        if color is None:
            color = self._icon_color

        self.set_icon(self._icons[index], size=size, colors=color)

    def next_icon(self):
        """
        Returns next icon of the current one and its index
        :return: tuple(QIcon, int)
        """

        new_index = max((self._icons.index(self._current_icon) + 1) % len(self._icons), 1)
        new_icon = self._icons[new_index]

        return new_icon, new_index


class ToolsetHelpWidget(base.BaseFrame, object):
    def __init__(self, parent=None):
        super(ToolsetHelpWidget, self).__init__(parent=parent)

    def ui(self):
        super(ToolsetHelpWidget, self).ui()

        self.setFrameStyle(QFrame.StyledPanel | QFrame.Plain)

        self._image_label = label.BaseLabel(parent=self)
        self._title_label = label.BaseLabel(parent=self).strong()
        self._description_text = QPlainTextEdit(parent=self)
        # self._description_text.setFrameShape(QFrame.NoFrame)
        self._description_text.setMinimumWidth(200)
        self._description_text.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self._gif_label = gif.GifLabel(parent=self)
        self._gif_label.set_size(256, 256)

        self.main_layout.addWidget(self._title_label)
        self.main_layout.addWidget(dividers.Divider())
        body_layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        body_layout.addWidget(self._description_text)
        body_layout.addWidget(self._gif_label)
        body_layout.addWidget(self._image_label)
        body_layout.addStretch()
        self.main_layout.addLayout(body_layout)

    def set_title(self, title):
        self._title_label.setText(str(title))

    def set_description(self, description):
        self._description_text.setPlainText(str(description))

    def set_image(self, image_file):
        if not image_file or not os.path.isfile(image_file):
            return
        self._image_label.setPixmap(QPixmap(image_file).scaled(QSize(256, 256), Qt.KeepAspectRatio))
        self._image_label.setVisible(True)
        self._gif_label.setVisible(False)

    def set_gif(self, gif_file):
        if not gif_file or not os.path.isfile(gif_file):
            return
        self._gif_label.set_file(gif_file)
        self._gif_label.setVisible(True)
        self._image_label.setVisible(False)


class ToolsetHelpEvent(base.BaseWidget, object):
    def __init__(self, layout, widgets=None, tooltip_widget=None):
        super(ToolsetHelpEvent, self).__init__()

        self._widgets = python.force_list(widgets)
        self._help_mode = False
        self._current_tooltip = ToolsetTooltipHelpDialog(self, tooltip_widget)
        self._current_widget = None
        self._temp_tooltip = None
        self.setVisible(False)

        layout.insertWidget(0, self)

        for widget in self._widgets:
            widget.installEventFilter(self)

    @property
    def help_mode(self):
        return self._help_mode

    @help_mode.setter
    def help_mode(self, flag):
        self._help_mode = flag

    @property
    def current_tooltip(self):
        return self._current_tooltip

    def eventFilter(self, obj, event):
        if event.type() == QEvent.Leave:
            if obj == self._current_widget:
                if self._temp_tooltip is not None:
                    self._current_widget.setToolTip(self._temp_tooltip)
                self._temp_tooltip = None
                self._current_widget = None

        pos = QCursor.pos()
        widget_under_cursor = QApplication.widgetAt(pos)
        # if not widget_under_cursor or widget_under_cursor.objectName() == '':
        if not widget_under_cursor:
            self._current_tooltip.hide()
            return super(ToolsetHelpEvent, self).eventFilter(obj, event)

        if self._help_mode:
            self._current_tooltip.move(pos.x() + 15, pos.y())
            if self._current_widget != widget_under_cursor:
                self._current_widget = widget_under_cursor
                tooltip_data = self._current_widget.property('tooltip_help')
                self._temp_tooltip = self._current_widget.toolTip()
                self._current_widget.setToolTip('')
                self._current_tooltip.update_tooltip(tooltip_data)
        else:
            self._current_tooltip.hide()

        return super(ToolsetHelpEvent, self).eventFilter(obj, event)


class ToolsetTooltipHelpDialog(QDialog, object):
    def __init__(self, parent, tooltip_widget):
        super(ToolsetTooltipHelpDialog, self).__init__(parent)

        self.setWindowFlags(Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint | Qt.Tool)
        self._tooltip_widget = tooltip_widget
        self._layout = layouts.VerticalLayout(margins=(0, 0, 0, 0))
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._tooltip_widget)
        self.setLayout(self._layout)
        self._tooltip_data = None
        # self.setStyleSheet('QDialog{background: rgb(60, 70, 75);}') CHANGED

    def update_tooltip(self, tooltip_data):
        self._tooltip_data = tooltip_data
        if not self._tooltip_data:
            self.hide()
            return

        title = tooltip_data.get('title', '')
        description = tooltip_data.get('description', '')
        image_file = tooltip_data.get('image', None)
        gif_file = tooltip_data.get('gif', None)

        self._tooltip_widget.set_title(title)
        self._tooltip_widget.set_description(description)
        self._tooltip_widget.set_image(image_file)
        self._tooltip_widget.set_gif(gif_file)

        self.show()
